/*
**   UltimateIRCd - Ultimate Internet Relay Chat Daemon, src/channel.c
**   Copyright (C) 1990 Jarkko Oikarinen and
**                      University of Oulu, Co Center
**
**   Rewritten for UltimateIRCd by ShadowRealm Creations.
**   Copyright (C) 1997-2000 Infomedia Inc.
**
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 1, or (at your option)
**   any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "channel.h"
#include "msg.h"	/* For TOK_*** and MSG_*** strings  */
#include "hash.h"	/* For CHANNELHASHSIZE */
#include "h.h"

#define MAXBOUNCE	5 /** Most sensible */

typedef struct ChLink CHLINK;
struct CHLink {
	aChannel *chan;
};

static int	bouncedtimes = 0;

struct CHLink	chlbounce[MAXBOUNCE];
int	chbounce = 0;


static long	opermode = 0;
aChannel *channel = NullChn;
extern aCRline *crlines;
extern char    *cannotjoin_msg;
// #define is_zombie(x,y) 0

static	void	add_invite PROTO((aClient *, aChannel *));
static	int	add_banid PROTO((aClient *, aChannel *, char *));
static	int	can_join PROTO((aClient *, aClient *, aChannel *, char *, char *, char **));
//static	int	can_join PROTO((aClient *, aClient *, aChannel *, char *));
static	int	channel_link PROTO((aClient *, aClient *, int, char **));
static	void	channel_modes PROTO((aClient *, char *, char *, aChannel *));
static	int	check_channelmask PROTO((aClient *, aClient *, char *));
static	int	del_banid PROTO((aChannel *, char *));
static	int	find_banid PROTO((aChannel *, char *));
/*extern	Link	*is_banned PROTO((aClient *, aChannel *));*/
static  int     have_ops PROTO((aChannel *));
//static	int	number_of_zombies PROTO((aChannel *));
static  int     is_deopped PROTO((aClient *, aChannel *));
static	void	set_mode PROTO((aChannel *, aClient *, int, char **, u_int *, \
			char [MAXMODEPARAMS][MODEBUFLEN+3], int));
static	void	make_mode_str PROTO((aChannel *, long, long, int,         \
			char [MAXMODEPARAMS][MODEBUFLEN+3], char *, char *, char));
static	int	do_mode_char PROTO((aChannel *, long, char, char *, u_int, \
			aClient *, u_int *, char [MAXMODEPARAMS][MODEBUFLEN+3], char));
static	void	do_mode PROTO((aChannel *, aClient *, aClient *, int, char **,
			int, int));
static	void	bounce_mode PROTO((aChannel *, aClient *, int, char **));

static	void	sub1_from_channel PROTO((aChannel *));

void	clean_channelname PROTO((char *));
void	del_invite PROTO((aClient *, aChannel *));

static	char	*PartFmt = ":%s PART %s";
static	char	*PartFmt2 = ":%s PART %s :%s";
static	long	linkedjoined = 0;

/*
 * some buffers for rebuilding channel/nick lists with ,'s
 */
static	char	nickbuf[BUFSIZE], buf[BUFSIZE];
static	char	modebuf[MODEBUFLEN], parabuf[MODEBUFLEN];

#ifdef USE_LONGMODE
typedef struct {
	long	mode;
	char	flag;
	int		halfop; /* 1 = yes 0 = no */
} aCtab;
// typedef struct CFlagTab aCtab;
#define MODESYS_LINKOK /* We do this for a TEST  */
aCtab cFlagTab[] = {
		{MODE_BAN,		'b', 1},
		{MODE_EXCEPT,		'e', 1},
		{MODE_FLOODLIMIT,	'f', 0}, /* Flood limiter */
		{MODE_HALFOP,		'h', 0},
		{MODE_INVITEONLY,	'i', 1},
		{MODE_KEY,		'k', 1},
		{MODE_LIMIT,		'l', 0},
		{MODE_MODERATED,	'm', 1},
		{MODE_NOPRIVMSGS,	'n', 1},
		{MODE_CHANOP,		'o', 0},
		{MODE_PRIVATE,		'p', 0},
		{MODE_RGSTR,		'r', 0},
		{MODE_SECRET,		's', 0},
		{MODE_TOPICLIMIT,	't', 1},
		{MODE_VOICE,		'v', 1},
		{MODE_NOCOLOR,		'x', 0},

		{MODE_ADMONLY,		'A', 0},
		{MODE_NOINVITE,		'I', 0}, /* no invites */	
		{MODE_NOKNOCK,	 	'K', 0}, /* knock knock (no way!) */
		{MODE_LINK,		'L', 0},
		{MODE_OPERONLY,		'O', 0},
		{MODE_RGSTRONLY,	'R', 0},
		{MODE_STRIP,		'S', 0}, /* works? */
        {0x0, 0x0, 0x0}
};
#endif


#define	BADOP_BOUNCE	1
#define	BADOP_USER	2
#define	BADOP_SERVER	3
#define	BADOP_OVERRIDE	4

char	cmodestring[512];

void	make_cmodestr(void)
{
	char	*p = &cmodestring[0];
	aCtab   *tab = &cFlagTab[0];
	
	while (tab->mode != 0x0) {
		*p = tab->flag;
		p++;
		tab++;
	}
	*p = '\0';
}


int	Halfop_mode(long mode) {
	aCtab	*tab = &cFlagTab[0];
	
	while (tab->mode != 0x0) {
		if (tab->mode == mode)
			return (tab->halfop == 1 ? TRUE : FALSE);
		tab++;
	}
}


/*
 * return the length (>=0) of a chain of links.
 */
static	int	list_length(lp)
Reg1	Link	*lp;
{
	Reg2	int	count = 0;

	for (; lp; lp = lp->next)
		count++;
	return count;
}

/*
** find_chasing
**	Find the client structure for a nick name (user) using history
**	mechanism if necessary. If the client is not found, an error
**	message (NO SUCH NICK) is generated. If the client was found
**	through the history, chasing will be 1 and otherwise 0.
*/
static	aClient *find_chasing(sptr, user, chasing)
aClient *sptr;
char	*user;
Reg1	int	*chasing;
{
	Reg2	aClient *who = find_client(user, (aClient *)NULL);

	if (chasing)
		*chasing = 0;
	if (who)
		return who;
	if (!(who = get_history(user, (long)KILLCHASETIMELIMIT)))
	    {
		sendto_one(sptr, err_str(ERR_NOSUCHNICK),
			   me.name, sptr->name, user);
		return NULL;
	    }
	if (chasing)
		*chasing = 1;
	return who;
}


/*
  Exception functions to work with mode +e
   -sts
*/

/* add_exbanid - add an id to be excepted to the channel bans  (belongs to cptr) */

static	int	add_exbanid(cptr, chptr, banid)
aClient	*cptr;
aChannel *chptr;
char	*banid;
{
	Reg1	Ban	*ban;
	Reg2	int	cnt = 0, len = 0;

	if (MyClient(cptr))
		(void)collapse(banid);
	for (ban = chptr->exlist; ban; ban = ban->next)
	    {
		len += strlen(ban->banstr);
		if (MyClient(cptr))
			if ((len > MAXBANLENGTH) || (++cnt >= MAXBANS))
			    {
				sendto_one(cptr, err_str(ERR_BANLISTFULL),
					   me.name, cptr->name,
					   chptr->chname, banid);
				return -1;
			    }
			else
			    {
				if (!match(ban->banstr, banid) ||
				    !match(banid, ban->banstr))
					return -1;
			    }
		else if (!mycmp(ban->banstr, banid))
			return -1;
		
	    }
	ban = make_ban();
	bzero((char *)ban, sizeof(Ban));
   /*	ban->flags = CHFL_BAN;			They're all bans!! */
	ban->next = chptr->exlist;
	ban->banstr = (char *)MyMalloc(strlen(banid)+1);
	(void)strcpy(ban->banstr, banid);
	ban->who = (char *)MyMalloc(strlen(cptr->name)+1);
	(void)strcpy(ban->who, cptr->name);
	ban->when = TStime();
	chptr->exlist = ban;
	return 0;
}
/*
 * del_exbanid - delete an id belonging to cptr
 */
static	int	del_exbanid(chptr, banid)
aChannel *chptr;
char	*banid;
{
	Reg1 Ban **ban;
	Reg2 Ban *tmp;

	if (!banid)
		return -1;
 	for (ban = &(chptr->exlist); *ban; ban = &((*ban)->next))
		if (mycmp(banid, (*ban)->banstr)==0)
		    {
			tmp = *ban;
			*ban = tmp->next;
			MyFree(tmp->banstr);
			MyFree(tmp->who);
			free_ban(tmp);
			return 0;
		    }
	return -1;
}

/*
 * find_banid - Find an exact match for a exban
 */
static	int	find_exbanid(chptr, banid)
aChannel *chptr;
char	*banid;
{
	Reg1 Ban **ban;
	Reg2 Ban *tmp;

	if (!banid)
		return -1;
	for (ban = &(chptr->exlist); *ban; ban = &((*ban)->next))
		if (!mycmp(banid, (*ban)->banstr)) return 1;
	return 0;
}

/*
 * Ban functions to work with mode +b
 */
/* add_banid - add an id to be banned to the channel  (belongs to cptr) */

static	int	add_banid(cptr, chptr, banid)
aClient	*cptr;
aChannel *chptr;
char	*banid;
{
	Reg1	Ban	*ban;
	Reg2	int	cnt = 0, len = 0;

	if (MyClient(cptr))
		(void)collapse(banid);
	for (ban = chptr->banlist; ban; ban = ban->next)
	    {
		len += strlen(ban->banstr);
		if (MyClient(cptr))
			if ((len > MAXBANLENGTH) || (++cnt >= MAXBANS))
			    {
				sendto_one(cptr, err_str(ERR_BANLISTFULL),
					   me.name, cptr->name,
					   chptr->chname, banid);
				return -1;
			    }
			else
			    {
				if (!match(ban->banstr, banid) ||
				    !match(banid, ban->banstr))
					return -1;
			    }
		else if (!mycmp(ban->banstr, banid))
			return -1;
		
	    }
	ban = make_ban();
	bzero((char *)ban, sizeof(Ban));
/*	ban->flags = CHFL_BAN;			They're all bans!! */
	ban->next = chptr->banlist;
	ban->banstr = (char *)MyMalloc(strlen(banid)+1);
	(void)strcpy(ban->banstr, banid);
	ban->who = (char *)MyMalloc(strlen(cptr->name)+1);
	(void)strcpy(ban->who, cptr->name);
	ban->when = TStime();
	chptr->banlist = ban;
	return 0;
}
/*
 * del_banid - delete an id belonging to cptr
 */
static	int	del_banid(chptr, banid)
aChannel *chptr;
char	*banid;
{
	Reg1 Ban **ban;
	Reg2 Ban *tmp;

	if (!banid)
		return -1;
 	for (ban = &(chptr->banlist); *ban; ban = &((*ban)->next))
		if (mycmp(banid, (*ban)->banstr)==0)
		    {
			tmp = *ban;
			*ban = tmp->next;
			MyFree(tmp->banstr);
			MyFree(tmp->who);
			free_ban(tmp);
			return 0;
		    }
	return -1;
}

/*
 * find_banid - Find an exact match for a ban
 */
static	int	find_banid(chptr, banid)
aChannel *chptr;
char	*banid;
{
	Reg1 Ban **ban;
	Reg2 Ban *tmp;

	if (!banid)
		return -1;
	for (ban = &(chptr->banlist); *ban; ban = &((*ban)->next))
		if (!mycmp(banid, (*ban)->banstr)) return 1;
	return 0;
}

/*
 * IsMember - returns 1 if a person is joined and not a zombie
 */
int	IsMember(cptr, chptr)
aClient *cptr;
aChannel *chptr;
{ Link *lp;
	/* Zombies have been removed */

	return (((lp=find_user_link(chptr->members, cptr)) /* &&
			!(lp->flags & CHFL_ZOMBIE)*/)?1:0);
}

/*
 * is_banned - returns a pointer to the ban structure if banned else NULL
 */
extern	Ban	*is_banned(cptr, sptr, chptr)
aClient *cptr, *sptr;
aChannel *chptr;
{
	Reg1	Ban	*tmp;


	char	*s, realhost[NICKLEN+USERLEN+HOSTLEN+6],
                    virthost[NICKLEN+USERLEN+HOSTLEN+6];

	int	dovirt = 0;

	if (!IsPerson(cptr))
		return NULL;

	if (IsAnOper(cptr)) return NULL;

        if (strcmp(cptr->user->host, cptr->user->virthost))
             dovirt = 1;

        s = make_nick_user_host(cptr->name, cptr->user->username,
                                  cptr->user->host);
        strcpy(realhost, s);

        s = make_nick_user_host(cptr->name, cptr->user->username,
                                  cptr->user->virthost);
        strcpy(virthost, s);

	for (tmp = chptr->exlist; tmp; tmp = tmp->next)
			if ((match(tmp->banstr, realhost) == 0) ||
					(dovirt && (match(tmp->banstr, virthost) == 0))
				)
						return (NULL); /* exception ! */


        for (tmp = chptr->banlist; tmp; tmp = tmp->next)
                if ((match(tmp->banstr, realhost) == 0) ||
                    (dovirt && (match(tmp->banstr, virthost) == 0))
                   )
                        break;
        return (tmp);

}

/*
 * adds a user to a channel by adding another link to the channels member
 * chain.
 */
static	void	add_user_to_channel(chptr, who, flags)
aChannel *chptr;
aClient *who;
int	flags;
{
	Reg1	Link *ptr;

	if (who->user)
	    {
		ptr = make_link();
		ptr->value.cptr = who;
		ptr->flags = flags;
		ptr->flood = (aFloodOpt *) MyMalloc(sizeof(aFloodOpt));
		ptr->flood->nmsg = 0;
		ptr->flood->lastmsg = 0;
		ptr->next = chptr->members;
		chptr->members = ptr;
		chptr->users++;

		ptr = make_link();
		ptr->value.chptr = chptr;
		ptr->next = who->user->channel;
		who->user->channel = ptr;
		who->user->joined++;
	    }
}

void	remove_user_from_channel(sptr, chptr)
aClient *sptr;
aChannel *chptr;
{
	Reg1	Link	**curr;
	Reg2	Link	*tmp;
	Reg3	Link	*lp = chptr->members;

	for (; lp && (lp->flags & CHFL_ZOMBIE || lp->value.cptr==sptr);
	    lp=lp->next);
	for (;;)
	{
	  for (curr = &chptr->members; (tmp = *curr); curr = &tmp->next)
		  if (tmp->value.cptr == sptr)
		      {
			  if (tmp->flood)
			  	MyFree((aFloodOpt *) tmp->flood);
			  *curr = tmp->next;
			  free_link(tmp);
			  break;
		      }
	  for (curr = &sptr->user->channel; (tmp = *curr); curr = &tmp->next)
		  if (tmp->value.chptr == chptr)
		      {
			  *curr = tmp->next;
			  free_link(tmp);
			  break;
		      }
	  sptr->user->joined--;
	  if (lp) break;
	  if (chptr->members) sptr = chptr->members->value.cptr;
	  else break;
	  sub1_from_channel(chptr);
	}
	sub1_from_channel(chptr);
}


static	int	have_ops(chptr)
aChannel *chptr;
{
	Reg1	Link	*lp;

	if (chptr)
        {
	  lp=chptr->members;
	  while (lp)
	  {
	    if (lp->flags & CHFL_CHANOP) return(1);
	    lp = lp->next;
	  }
        }
	return 0;
}

int	is_chan_op(cptr, chptr)
aClient *cptr;
aChannel *chptr;
{
	Reg1	Link	*lp;

	if (chptr)
		if ((lp = find_user_link(chptr->members, cptr)))
			return ((lp->flags & CHFL_CHANOP));

	return 0;
}


/* This was the original function that allowed mode hacking - not
   anymore. -- Barubary */
static	int	is_deopped(cptr, chptr)
aClient *cptr;
aChannel *chptr;
{
	if (!IsPerson(cptr)) return 0;
	return !is_chan_op(cptr, chptr);
}



int	is_zombie(cptr, chptr)
aClient *cptr;
aChannel *chptr;
{
	Reg1	Link	*lp;

/* We have no Zombies...
	if (chptr)
		if ((lp = find_user_link(chptr->members, cptr)))
			return (lp->flags & CHFL_ZOMBIE);
*/
	return 0;
}

int	has_voice(cptr, chptr)
aClient *cptr;
aChannel *chptr;
{
	Reg1	Link	*lp;

	if (chptr)
		if ((lp = find_user_link(chptr->members, cptr)) &&
		    !(lp->flags & CHFL_ZOMBIE))
			return (lp->flags & CHFL_VOICE);

	return 0;
}

int		is_halfop(cptr, chptr)
aClient *cptr;
aChannel *chptr;
{
        Reg1    Link    *lp;

        if (chptr)
                if (lp = find_user_link(chptr->members, cptr))
                 	if (!is_chan_op(cptr, chptr)) /* excessive but needed */
                        return (lp->flags & CHFL_HALFOP);

        return 0;
}

#define CANNOT_SEND_MODERATED 1
#define CANNOT_SEND_NOPRIVMSGS 2
#define CANNOT_SEND_NOCOLOR 3
#define CANNOT_SEND_BAN 4
// #define CAN_SEND_STRIP 5

int	can_send(cptr, chptr, msgtext)
aClient *cptr;
aChannel *chptr;
char	*msgtext;
{
	Reg1	Link	*lp;
	Reg2	int	member;

	/* Moved check here, kinda faster.
	 * Note IsULine only uses the other parameter. -Donwulff */
	if (IsULine(cptr,cptr)||IsServer(cptr)) return 0;

	member = IsMember(cptr, chptr);
	lp = find_user_link(chptr->members, cptr);

	if (IsSkoAdmin(cptr)) return 0;

	if (chptr->mode.mode & MODE_MODERATED &&
	    (!lp || !(lp->flags & (CHFL_CHANOP|CHFL_HALFOP|CHFL_VOICE)) ||
	    (lp->flags & CHFL_ZOMBIE)))
			return (MODE_MODERATED);

	if (chptr->mode.mode & MODE_NOPRIVMSGS && !member)
		return (MODE_NOPRIVMSGS);

        if (chptr->mode.mode & MODE_NOCOLOR)
          if(strchr(msgtext, 3) ||
             strchr(msgtext, 27))

            return (CANNOT_SEND_NOCOLOR);
/* Fixed strip bug --sts */
/*	if (chptr->mode.mode & MODE_STRIP) {
		if ((strchr(msgtext, 3) != NULL) ||
                 (strchr(msgtext, 27) != NULL)) {
			return (CAN_SEND_STRIP);			
		}
	}	
*/


if ((!lp || !(lp->flags & (CHFL_CHANOP|CHFL_HALFOP|CHFL_VOICE)) ||
    (lp->flags & CHFL_ZOMBIE)) && MyClient(cptr) &&
    is_banned(cptr, cptr, chptr))
	return (MODE_BAN);

	return 0;
}

aChannel *find_channel(chname, chptr)
Reg1	char	*chname;
Reg2	aChannel *chptr;
{
	return hash_find_channel(chname, chptr);
}


/*
 * write the "simple" list of channel modes for channel chptr onto buffer mbuf
 * with the parameters in pbuf.
 */
static	void	channel_modes(cptr, mbuf, pbuf, chptr)
aClient	*cptr;
Reg1	char	*mbuf, *pbuf;
aChannel *chptr;
{
	int				*ip, *ipx;
	long				zode;
	aCtab			*tab = &cFlagTab[0];
	char				bcbuf[1024];

	*mbuf++ = '+';
	while (tab->mode != 0x0) {
	  if ((chptr->mode.mode & tab->mode)) {
		zode = chptr->mode.mode;
		if (!(zode & (MODE_LIMIT|MODE_KEY|MODE_LINK)))
			if (!(zode & (MODE_CHANOP|MODE_VOICE|MODE_HALFOP)))
				if (!(zode & (MODE_EXCEPT|MODE_BAN)))
						*mbuf++ = tab->flag;
	   }
	   tab++;
	}
	if (chptr->mode.limit)
	    {
		*mbuf++ = 'l';
		if (IsMember(cptr, chptr) || IsServer(cptr) || IsULine(cptr,cptr) || IsAnOper(cptr))
			(void)ircsprintf(pbuf, "%d ", chptr->mode.limit);
	    }
	if (*chptr->mode.key)
	    {
		*mbuf++ = 'k';
		if (IsMember(cptr, chptr) || IsServer(cptr) || IsULine(cptr,cptr) || IsAnOper(cptr))
			(void)strcat(pbuf, chptr->mode.key);
	    }
        if (*chptr->mode.link)
            {
                *mbuf++ = 'L';
                if (IsMember(cptr, chptr) || IsServer(cptr) || IsULine(cptr,cptr))
                        (void)strcat(pbuf, chptr->mode.link);
            }
	if (chptr->mode.per)
	{
		*mbuf++ = 'f';
                if (IsMember(cptr, chptr) || IsServer(cptr) || IsULine(cptr,cptr))
                {
			if (chptr->mode.kmode == 1)
				ircsprintf(bcbuf, "*%i:%i", chptr->mode.msgs, chptr->mode.per);
			else
				ircsprintf(bcbuf, "%i:%i", chptr->mode.msgs, chptr->mode.per);
                	(void)strcat(pbuf, bcbuf);
                }

	}

	*mbuf++ = '\0';
	return;
}


static	int send_mode_list(cptr, chname, creationtime, top, mask, flag)
aClient	*cptr;
Link	*top;
int	mask;
char	flag, *chname;
time_t	creationtime;
{
	Reg1	Link	*lp;
	Reg2	char	*cp, *name;
	int	count = 0, send = 0, sent = 0;

	cp = modebuf + strlen(modebuf);
	if (*parabuf)	/* mode +l or +k xx */
		count = 1;
	for (lp = top; lp; lp = lp->next)
	    {
		/* 
		 * Okay, since ban's are stored in their own linked
		 * list, we won't even bother to check if CHFL_BAN
		 * is set in the flags. This should work as long
		 * as only ban-lists are feed in with CHFL_BAN mask.
		 * However, we still need to typecast... -Donwulff 
		 */
		if ((mask == CHFL_BAN) || (mask == CHFL_EXCEPT)) {
/*			if (!(((Ban *)lp)->flags & mask)) continue; */
			name = ((Ban *)lp)->banstr;
		} else {
			if (!(lp->flags & mask)) continue;
			name = lp->value.cptr->name;
		}
		if (strlen(parabuf) + strlen(name) + 11 < (size_t) MODEBUFLEN)
		    {
			if(*parabuf) (void)strcat(parabuf, " ");
			(void)strcat(parabuf, name);
			count++;
			*cp++ = flag;
			*cp = '\0';
		    }
		else if (*parabuf)
			send = 1;
		if (count == RESYNCMODES)
			send = 1;
		if (send)
		    {
		       /* cptr is always a server! So we send creationtimes */
			sendmodeto_one(cptr, me.name, chname, modebuf,
				   parabuf, creationtime);
                        sent = 1;
			send = 0;
			*parabuf = '\0';
			cp = modebuf;
			*cp++ = '+';
			if (count != RESYNCMODES)
			    {
				(void)strcpy(parabuf, name);
				*cp++ = flag;
			    }
			count = 0;
			*cp = '\0';
		    }
	    }
     return sent;
}

/*
 * send "cptr" a full list of the modes for channel chptr.
 */
void	send_channel_modes(cptr, chptr)
aClient *cptr;
aChannel *chptr;
{       int sent;
	if (*chptr->chname != '#')
		return;

	*parabuf = '\0';
	*modebuf = '\0';
	channel_modes(cptr, modebuf, parabuf, chptr);
	sent=send_mode_list(cptr, chptr->chname, chptr->creationtime,
	    chptr->members, CHFL_CHANOP, 'o');
	if (!sent && chptr->creationtime)
	  sendto_one(cptr, ":%s %s %s %s %s %lu", me.name,
	      (IsToken(cptr)?TOK_MODE:MSG_MODE), chptr->chname, modebuf,
	      parabuf, chptr->creationtime);
	else if (modebuf[1] || *parabuf)
	  sendmodeto_one(cptr, me.name,
	      chptr->chname, modebuf, parabuf, chptr->creationtime);

	*parabuf = '\0';
	*modebuf = '+';
	modebuf[1] = '\0';
	sent=send_mode_list(cptr, chptr->chname, chptr->creationtime,
	    chptr->members, CHFL_HALFOP, 'h');
	if (!sent && chptr->creationtime)
	  sendto_one(cptr, ":%s %s %s %s %s %lu", me.name,
	      (IsToken(cptr)?TOK_MODE:MSG_MODE), chptr->chname, modebuf,
	      parabuf, chptr->creationtime);
	else if (modebuf[1] || *parabuf)
	  sendmodeto_one(cptr, me.name,
	      chptr->chname, modebuf, parabuf, chptr->creationtime);
	
	*parabuf = '\0';
	*modebuf = '+';
	modebuf[1] = '\0';
	(void)send_mode_list(cptr, chptr->chname,chptr->creationtime,
	    chptr->banlist, CHFL_BAN, 'b');
	if (modebuf[1] || *parabuf)
	  sendmodeto_one(cptr, me.name, chptr->chname, modebuf,
	      parabuf, chptr->creationtime);

	*parabuf = '\0';
	*modebuf = '+';
	modebuf[1] = '\0';
	(void)send_mode_list(cptr, chptr->chname,chptr->creationtime,
	    chptr->exlist, CHFL_EXCEPT, 'e');
	if (modebuf[1] || *parabuf)
	  sendmodeto_one(cptr, me.name, chptr->chname, modebuf,
	      parabuf, chptr->creationtime);

	*parabuf = '\0';
	*modebuf = '+';
	modebuf[1] = '\0';
	(void)send_mode_list(cptr, chptr->chname,chptr->creationtime,
	    chptr->members, CHFL_VOICE, 'v');
	if (modebuf[1] || *parabuf)
	  sendmodeto_one(cptr, me.name, chptr->chname, modebuf,
	      parabuf, chptr->creationtime);
}

/*********/

/*
 * m_samode
 * parv[0] = sender
 * parv[1] = channel
 * parv[2] = modes
 * -taz
 */
int	m_samode(cptr, sptr, parc, parv)
aClient *cptr, *sptr;
int	parc;
char	*parv[];
{
	static char tmp[MODEBUFLEN];
	int badop, sendts;
	aChannel *chptr;

	if (check_registered(cptr))
		return 0;

	if (!IsPrivileged(cptr) || !IsServicesOper(sptr))
	{
		sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
		return 0;
	}

	if (parc > 2)
	{
		chptr = find_channel(parv[1], NullChn);
		if (chptr == NullChn)
			return 0;
	}
	else
	{
		sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS),
			   me.name, parv[0], "MODE");
 	 	return 0;
	}
	opermode = 0; 
	(void)do_mode(chptr, cptr, sptr, parc - 2, parv + 2, 0, 1);

	return 0;

}


/*
 * m_mode -- written by binary (garryb@binary.islesfan.net)
 *	Completely rewrote it.  The old mode command was 820 lines of ICKY
 * coding, which is a complete waste, because I wrote it in 570 lines of
 * *decent* coding.  This is also easier to read, change, and fine-tune.  Plus,
 * everything isn't scattered; everything's grouped where it should be.
 *
 * parv[0] - sender
 * parv[1] - channel
 */
int	m_mode(cptr, sptr, parc, parv)
aClient *cptr;
aClient *sptr;
int	parc;
char	*parv[];
{
	static char tmp[MODEBUFLEN];
	static char mode_buf[MODEBUFLEN], parabuf[MODEBUFLEN];
	char	tmp1[1024];
//	int	opermode = 0;
	
	long unsigned timestamp, sendts;
	Ban *ban;
	Link *l;
	aChannel *chptr;

	if (check_registered(sptr))
		return 0;

	/* Now, try to find the channel in question */
	if (parc > 1)
	{
		chptr = find_channel(parv[1], NullChn);
		if (chptr == NullChn)
			return m_umode(cptr, sptr, parc, parv);
	}
	else
	{
		sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS),
			   me.name, parv[0], "MODE");
 	 	return 0;
	}

	sptr->flags&=~FLAGS_TS8;

	if (MyConnect(sptr))
		clean_channelname(parv[1]);
	if (check_channelmask(sptr, cptr, parv[1]))
		return 0;

	if (parc < 3)
	{
		*mode_buf = *parabuf = '\0';
		mode_buf[1] = '\0';
		channel_modes(sptr, mode_buf, parabuf, chptr);
		sendto_one(sptr, rpl_str(RPL_CHANNELMODEIS), me.name, parv[0],
			   chptr->chname, mode_buf, parabuf);
		sendto_one(sptr, rpl_str(RPL_CREATIONTIME), me.name, parv[0],
				 chptr->chname, chptr->creationtime);
		return 0;
	}

	if (IsPerson(sptr) && parc<4 && ((*parv[2]=='b' && parv[2][1]=='\0')
		|| (parv[2][1]=='b' && parv[2][2] == '\0' && (*parv[2] == '+'
		|| *parv[2] == '-'))))
	{
		/* send ban list */
		for (ban=chptr->banlist; ban; ban=ban->next)
			sendto_one(sptr, rpl_str(RPL_BANLIST), me.name,
				sptr->name, chptr->chname, ban->banstr,
				ban->who, ban->when);
		sendto_one(cptr,
			rpl_str(RPL_ENDOFBANLIST), me.name, sptr->name,
			chptr->chname);
		return 0;
	}
	if (IsPerson(sptr) && parc<4 && ((*parv[2]=='e' && parv[2][1]=='\0')
		|| (parv[2][1]=='e' && parv[2][2] == '\0' && (*parv[2] == '+'
		|| *parv[2] == '-'))))
	{
		/* send exban list */
		for (ban=chptr->exlist; ban; ban=ban->next)
			sendto_one(sptr, rpl_str(RPL_EXLIST), me.name,
				sptr->name, chptr->chname, ban->banstr,
				ban->who, ban->when);
		sendto_one(cptr,
			rpl_str(RPL_ENDOFEXLIST), me.name, sptr->name,
			chptr->chname);
		return 0;
	}
	opermode = 0;
	/* opermode stuff --sts */
	if (IsPerson(sptr) && !IsULine(cptr, sptr) && !is_chan_op(sptr, chptr))
		if (IsServicesOper(sptr) || IsSkoAdmin(sptr)) {
			sendts = 0;
			opermode = 1;
			goto aftercheck;
		}
/*-------*/


	if (is_half_op(sptr, chptr)) {
		opermode = 2;
		goto aftercheck;
	}
	if (IsPerson(sptr) && !IsULine(cptr, sptr) && !is_chan_op(sptr, chptr)
		&& !is_half_op(sptr, chptr)
		&& (cptr == sptr || !IsServicesOper(sptr) || !IsOper(sptr)))
	{	
		if (cptr == sptr)
		{
			sendto_one(sptr, err_str(ERR_CHANOPRIVSNEEDED),
				me.name, parv[0], chptr->chname);
			return 0;
		}
		sendto_one(cptr, ":%s MODE %s -oh %s %s 0",
			me.name, chptr->chname, parv[0]);
			/* Tell the other server that the user is
			 * de-opped.  Fix op desyncs. */
		bounce_mode(chptr, cptr, parc - 2, parv + 2);
		return 0;
	}
	
	if (IsServer(sptr) && (sendts = atoi(parv[parc-1]))
		&& !IsULine(cptr,sptr) && chptr->creationtime
		&& sendts > chptr->creationtime)
	{
		if (!(*parv[2] == '&')) /* & denotes a bounce */
		{
            /* !!! */
			sendto_umode(UMODE_OPER, "*** TS bounce for %s - %lu(ours) %lu(theirs)",
				 chptr->chname, chptr->creationtime, sendts);
			bounce_mode(chptr, cptr, parc - 2, parv + 2);
			return 0;
		}
		/* other server will resync soon enough... */
	}
	if (IsServer(sptr) && !sendts && *parv[parc - 1] != '0')
		sendts = -1;
	if (IsServer(sptr) && sendts != -1)
		parc--; /* server supplied a time stamp, remove it now */

aftercheck:
/*	if (IsPerson(sptr) && IsServicesOper(sptr)) {
		if (!is_chan_op(sptr, chptr)) {
			if (MyClient(sptr) && !IsULine(cptr, cptr))
				sendto_umode(UMODE_OPER, "*** OperMode [IRCop: %s] - [Channel: %s] - [Mode: %s %s]",
        	 		   sptr->name, chptr->chname, mode_buf, parabuf);
			sendts = 0;
		}
	}	
*/
	/* Filter out the unprivileged FIRST. *
	 * Now, we can actually do the mode.  */

	(void)do_mode(chptr, cptr, sptr, parc - 2, parv + 2, sendts, 0);
        
	return 0;
}

/* bounce_mode -- written by binary
 *	User or server is NOT authorized to change the mode.  This takes care
 * of making the bounce string and bounce it.  Because of the 1 for the bounce
 * param (last param) of the calls to set_mode and make_mode_str, it will not
 * set the mode, but create the bounce string.
 */
void	bounce_mode(chptr, cptr, parc, parv)
aChannel	*chptr;
aClient		*cptr;
int		parc;
char		*parv[];
{
	char	pvar[MAXMODEPARAMS][MODEBUFLEN+3];
	int	pcount;
	long	oldm, oldl;
	char	mode_buf[MODEBUFLEN], parabuf[MODEBUFLEN];
	
	oldm = chptr->mode.mode;
	oldl = chptr->mode.limit;
	set_mode(chptr, cptr, parc, parv, &pcount, pvar, 1);
	make_mode_str(chptr, oldm, oldl, pcount, pvar, mode_buf,
		parabuf, 1);
	
	sendto_one(cptr, ":%s MODE %s &%s %s %lu", me.name, chptr->chname,
		mode_buf, parabuf, chptr->creationtime);
	/* the '&' denotes a bounce so servers won't bounce a bounce */
}

/* do_mode -- written by binary
 *	User or server is authorized to do the mode.  This takes care of
 * setting the mode and relaying it to other users and servers.
 */
void	do_mode(chptr, cptr, sptr, parc, parv, sendts, samode)
aChannel	*chptr;
aClient		*cptr, *sptr;
int		parc, sendts, samode;
char		*parv[];
{
	char	pvar[MAXMODEPARAMS][MODEBUFLEN+3];
	int	pcount;
	long	oldm, oldl;
	char	tschange = 0, isbounce = 0; /* fwd'ing bounce */
	char	mode_buf[MODEBUFLEN], parabuf[MODEBUFLEN];

	if (**parv == '&')
		isbounce = 1;
	oldm = chptr->mode.mode;
	oldl = chptr->mode.limit;

	set_mode(chptr, sptr, parc, parv, &pcount, pvar, 0);
	make_mode_str(chptr, oldm, oldl, pcount, pvar, mode_buf,
		parabuf, 0);

	if (IsServer(sptr))
	{
		if (sendts > 0)
		{
			if (!chptr->creationtime || sendts < chptr->creationtime)
			{
				tschange = 1;
				if (chptr->creationtime != 0)
					sendto_umode(UMODE_OPER, "*** TS fix for %s - %lu(ours) %lu(theirs)",
					chptr->chname, chptr->creationtime, sendts);				
				chptr->creationtime = sendts;
				/* new chan or our timestamp is wrong */
				/* now works for double-bounce prevention */

			}
			if (sendts > chptr->creationtime && chptr->creationtime)
			{
				/* theirs is wrong but we let it pass anyway */
				sendts = chptr->creationtime;
				sendto_one(cptr, ":%s MODE %s + %lu", me.name,
					chptr->chname, chptr->creationtime);
			}
		}
		if (sendts == -1 && chptr->creationtime)
			sendts = chptr->creationtime;
	}
	if (*mode_buf == '\0' || (*(mode_buf+1) == '\0' && (*mode_buf == '+' ||
			*mode_buf == '-')))
	{
		if (tschange || isbounce) /* relay bounce time changes */
			sendto_serv_butone(cptr, ":%s MODE %s %s+ %lu", me.name,
				chptr->chname, isbounce ? "&" : "",
				chptr->creationtime);
		return; /* nothing to send */
	}
	/* opermode for twimodesystem --sts*/

/*	if (opermode == 1) */
	if (IsAdmin(sptr))
	 {
		if (MyClient(sptr))
/*
			sendto_umode(UMODE_OPER, "*** OperMode [IRCop: %s] - [Channel: %s] - [Mode: %s %s]",
       			   sptr->name, chptr->chname, mode_buf, parabuf);
*/
		sendts = 0;
	}
	
    if (IsPerson(sptr) && samode && MyClient(sptr))
    {
	     	sendto_serv_butone(NULL,
	     		":%s GLOBOPS :%s used SAMODE %s (%s %s)",
			me.name, sptr->name, chptr->chname, mode_buf, parabuf);
		sendto_failops_whoare_opers(
			"from %s: %s used SAMODE %s (%s %s)", 
			me.name, sptr->name, chptr->chname, mode_buf, parabuf);
		sptr = &me;
		sendts = 0;
	}
	sendto_channel_butserv(chptr, sptr, ":%s MODE %s %s %s",
       		sptr->name, chptr->chname, mode_buf, parabuf);
       	if (IsServer(sptr) && sendts != -1)
		sendto_serv_butone(cptr, ":%s MODE %s %s%s %s %lu",
          		sptr->name, chptr->chname, isbounce ? "&" : "",
          		mode_buf, parabuf, sendts);
        else
        	sendto_serv_butone(cptr, ":%s MODE %s %s%s %s NoTS",
          		sptr->name, chptr->chname, isbounce ? "&" : "",
          		mode_buf, parabuf);
          	/* tell them it's not a timestamp, in case the last param
          	** is a number. */
}
/* make_mode_str -- written by binary
 *	Reconstructs the mode string, to make it look clean.  mode_buf will
 *  contain the +x-y stuff, and the parabuf will contain the parameters.
 *  If bounce is set to 1, it will make the string it needs for a bounce.
 */
void	make_mode_str(chptr, oldm, oldl, pcount, pvar, mode_buf, parabuf, bounce)
aChannel	*chptr;
int	pcount;
long	oldm, oldl;
char pvar[MAXMODEPARAMS][MODEBUFLEN+3], *mode_buf, *parabuf, bounce;
{

	char	*x = mode_buf, tmpbuf[MODEBUFLEN], *tmpstr, foo;
	// int	*ip;
	aCtab *tab = &cFlagTab[0];
	int	what, cnt, z;
	
	what = 0;
	
	*tmpbuf = '\0';
	*mode_buf = '\0';
	*parabuf = '\0';
	what = 0;
	/* + param-less modes */
	tab = &cFlagTab[0];
	while (tab->mode != 0x0) {
		if (chptr->mode.mode & tab->mode) {
			if (!(oldm & tab->mode)) {
				if (what != MODE_ADD)
				{
					*x++ = bounce ? '-' : '+';
					what = MODE_ADD;
				}	
				*x++ = tab->flag;
			}
		}
		tab++;
	}
	
	*x = '\0';
/* + param-less modes */
	tab = &cFlagTab[0];
	while (tab->mode != 0x0) {
		if (!(chptr->mode.mode & tab->mode)) {
			if (oldm & tab->mode) {
				if (what != MODE_DEL)
				{
					*x++ = bounce ? '+' : '-';
					what = MODE_DEL;
				}	
				*x++ = tab->flag;
			}
		}
		tab++;
	}
	
	*x = '\0';
	/* reconstruct bkov chain */
	for (cnt = 0; cnt < pcount; cnt++)
	{
		if ((*(pvar[cnt]) == '+') && what != MODE_ADD)
		{
			*x++ = bounce ? '-' : '+';
			what = MODE_ADD;
		}
		if ((*(pvar[cnt]) == '-') && what != MODE_DEL)
		{
			*x++ = bounce ? '+' : '-';
			what = MODE_DEL;
		}
		*x++ = *(pvar[cnt] + 1);
		tmpstr = &pvar[cnt][2];
		strncat(parabuf, tmpstr, MODEBUFLEN - 1);
		parabuf[MODEBUFLEN-1] = '\0';
		z = strlen(parabuf);
		parabuf[z] = ' ';	/* add a space */
		parabuf[z+1] = '\0';
	}
	/* user limit */
	if (chptr->mode.limit != oldl)
	{
		if ((!bounce && chptr->mode.limit == 0) ||
			(bounce && chptr->mode.limit != 0))
		{
			if (what != MODE_DEL)
			{
				*x++ = '-';
				what = MODE_DEL;
			}
			if (bounce)
				chptr->mode.limit = 0; /* set it back */
			*x++ = 'l';
		}
		else
		{
			if (what != MODE_ADD)
			{
				*x++ = '+';
				what = MODE_ADD;
			}
			*x++ = 'l';
			if (bounce)
				chptr->mode.limit = oldl; /* set it back */
			ircsprintf(parabuf, "%s%d", parabuf, chptr->mode.limit);
		}
	}
	if (bounce)
		chptr->mode.mode = oldm;
	z = strlen(parabuf);
	if (parabuf[z-1] == ' ')
		parabuf[z-1] = '\0';
	*x = '\0';
	if (*mode_buf == '\0')
	{
		*mode_buf = '+';
		mode_buf++;
		*mode_buf = '\0';
		/* Don't send empty lines. */
	}
	return;
}


/* do_mode_char
 *  processes one mode character
 *  returns 1 if it ate up a param, otherwise 0
 *	written by binary
 *  modified for Unreal by stskeeps..
 */

int do_mode_char(chptr, modetype, modechar, param, what, cptr, pcount, pvar,
	bounce)
char *param, modechar, pvar[MAXMODEPARAMS][MODEBUFLEN+3], bounce;
u_int what;
long modetype;
u_int *pcount;
aClient *cptr;
aChannel	*chptr;

{
	aCtab 	*tab = &cFlagTab[0];
	

	int retval = 0;
	Link	*member;
	aClient	*who;
	unsigned int tmp = 0;
	char tmpbuf[256], *tmpstr;
	char tc; /* */
	int chasing, x;
	int xxi, xyi, xzi;
	char *xxx;
	char *xp;
	
	chasing = 0;
	if (opermode == 2 && !is_chan_op(cptr, chptr)) {
	 /* Ugly halfop hack --sts 
	    - this allows halfops to do +b +e +v and so on */
	  if (Halfop_mode(modetype) == FALSE) {
			while (tab->mode != 0x0) {
				if (tab->mode == modetype) {
					sendto_one(cptr, err_str(ERR_NOTFORHALFOPS),
						me.name,
						cptr->name,						
						tab->flag);
				}
				tab++;
			}
	 		return(0);
	  }
	}
	switch(modetype)
	{
		case MODE_RGSTR:
			if (IsULine(cptr,cptr) || IsServer(cptr)) 
			{
				if (what == MODE_ADD)
				chptr->mode.mode |= MODE_RGSTR;
				else
				chptr->mode.mode &= ~MODE_RGSTR;
			}else 
			{
	    			sendto_one(cptr, err_str(ERR_ONLYSERVERSCANCHANGE),
				    	me.name, cptr->name, chptr->chname);
			}
				break;
		case MODE_OPERONLY:
			if (!IsAnOper(cptr) && !IsServer(cptr)) {
				sendto_one(cptr, ":%s NOTICE %s :*** Only IRCops can set that mode", me.name, cptr->name);
				break;
			}
		case MODE_ADMONLY:
			if (!IsSkoAdmin(cptr) && !IsServer(cptr)) {
				sendto_one(cptr, ":%s NOTICE %s :*** Only admins can set that mode", me.name, cptr->name);
				break;
			}
		case MODE_SECRET:
	        case MODE_PRIVATE:
		case MODE_MODERATED:
		case MODE_TOPICLIMIT:
		case MODE_NOPRIVMSGS:
		case MODE_INVITEONLY:
		case MODE_RGSTRONLY:
		case MODE_NOCOLOR:
		case MODE_STRIP:
		case MODE_NOKNOCK:
		case MODE_NOINVITE:
			/* +spi bugfix.. */
			if (modetype == MODE_SECRET && (chptr->mode.mode & MODE_PRIVATE))
				chptr->mode.mode &= ~MODE_PRIVATE;
			if (modetype == MODE_PRIVATE && (chptr->mode.mode & MODE_SECRET))
				chptr->mode.mode &= ~MODE_SECRET;
			if (modetype == MODE_NOCOLOR && (chptr->mode.mode & MODE_STRIP))
				chptr->mode.mode &= ~MODE_STRIP;
			if (modetype == MODE_STRIP && (chptr->mode.mode & MODE_NOCOLOR))
				chptr->mode.mode &= ~MODE_NOCOLOR;
			retval = 0;
			if (what == MODE_ADD)
				chptr->mode.mode |= modetype;
			else
				chptr->mode.mode &= ~modetype;
			break;


/* do pro-opping here (popping) */
		case MODE_HALFOP:
		case MODE_CHANOP:
		case MODE_VOICE:
			switch (modetype)
			{
				case MODE_HALFOP: xxx = "dehalfop"; break;
				case MODE_CHANOP: xxx = "deop"; break;
				case MODE_VOICE: xxx ="devoice"; break;
				default: xxx = "whattheheckisthatmode?";
			}
			if (!param || *pcount >= MAXMODEPARAMS)
			{
				retval = 0;
				break;
			}
			retval = 1;
			if (!(who = find_chasing(cptr, param, &chasing)))
				break;
	  		if (!(member = find_user_link(chptr->members, who)))
			{
	    			sendto_one(cptr, err_str(ERR_USERNOTINCHANNEL),
					   me.name, cptr->name,
					   who->name, chptr->chname);
				break;
			}

			/* if the user is +p, prevent a deop/DeVoice from local user */
			if (!IsServer(cptr) && !IsULine(cptr,cptr) && !IsSuper(cptr)) {
				if (what == MODE_DEL && (who->umodes & UMODE_SUPER)  && MyClient(cptr)) {
                          sendto_one(cptr,":%s NOTICE %s :%s cannot be deopped/devoiced in %s (Protected IRC Operator)",
				  me.name, cptr->name, who->name, chptr->chname);
			sendto_netglobal("Overriding DeOp/DeVoice for %s on channel %s made by: %s (%s@%s)",
			          who->name, chptr->chname, cptr->name, cptr->user->username, cptr->user->host);
		    sendto_serv_butone(&me, ":%s NETG :Overriding DeOp/DeVoice for %s on channel %s made by: %s (%s@%s)",
			          me.name, who->name, chptr->chname, cptr->name, cptr->user->username, cptr->user->host);

			  break;
			}
			}

			/* Lets protect +S clients */

			if (!IsServer(cptr) && !IsULine(cptr,cptr) && !IsServices(cptr)) {
				if (what == MODE_DEL && (who->umodes & UMODE_SERVICES)  && MyClient(cptr)) {
                     sendto_one(cptr, ":%s NOTICE %s :*** [%s] Cannot kill, deop, or kick Services Clients.", me.name, cptr->name, who->name);
			  break;
			}
			}

			tmp = member->flags;
			if (what == MODE_ADD)
				member->flags |= modetype;
			else
				member->flags &= ~modetype;
			if (tmp == member->flags &&
					(bounce || !IsULine(cptr,cptr)))
				break;
			/* It's easier to undo the mode here instead of later
			 * when you call make_mode_str for a bounce string.
			 * Why set it if it will be instantly removed?
			 * Besides, pvar keeps a log of it. */
			if (bounce)
				member->flags = tmp;
			if (modetype == MODE_CHANOP)
				tc = 'o';
			if (modetype == MODE_HALFOP)
				tc = 'h';
			if (modetype == MODE_VOICE)
				tc = 'v';
					
			(void)ircsprintf(pvar[*pcount], "%c%c%s",
				what == MODE_ADD ? '+' : '-',
				tc,
				who->name);

			(*pcount)++;
			break;
		case MODE_LIMIT:
			if (what == MODE_ADD)
			{
				if (!param)
				{
					retval = 0;
					break;
				}
				retval = 1;
				tmp = atoi(param);
				if (chptr->mode.limit == tmp)
					break;
				chptr->mode.limit = tmp;
			}
			else
			{
				retval = 0;
				if (!chptr->mode.limit)
					break;

				strcpy(chptr->mode.link, "");
			(void)ircsprintf(pvar[*pcount], "%cL", '-');
			(*pcount)++;				

				chptr->mode.limit = 0;
			}
			break;
		case MODE_KEY:
			if (!param || *pcount >= MAXMODEPARAMS)
			{
				retval = 0;
				break;
			}
			retval = 1;
			for (x = 0; x < *pcount; x++)
			{
				if (pvar[x][1] == 'k')
				{	/* don't allow user to change key
					 * more than once per command. */
					retval = 0;
					break;
				}
			}
			if (retval == 0) /* you can't break a case from loop */
				break;
			if (what == MODE_ADD)
			{
				if (!bounce) /* don't do the mode at all. */
					strncpyzt(chptr->mode.key, param,
						sizeof(chptr->mode.key));
				tmpstr = param;
			}
			else
			{
				if (!*chptr->mode.key)
					break; /* no change */
				strncpyzt(tmpbuf, chptr->mode.key,
					sizeof(tmpbuf));
				tmpstr = tmpbuf;
				if (!bounce)
					strcpy(chptr->mode.key, "");
			}
			retval = 1;
			
			(void)ircsprintf(pvar[*pcount], "%ck%s",
				what == MODE_ADD ? '+' : '-', tmpstr);
			(*pcount)++;				
			break;
		
		case MODE_BAN:
			if (!param || *pcount >= MAXMODEPARAMS)
			{
				retval = 0;
				break;
			}
			retval = 1;
			tmpstr = pretty_mask(param);
			/* For bounce, we don't really need to worry whether
			 * or not it exists on our server.  We'll just always
			 * bounce it. */
			if (!bounce &&
			   ((what==MODE_ADD && add_banid(cptr, chptr, tmpstr))
			   || (what == MODE_DEL && del_banid(chptr, tmpstr))))
				break; /* already exists */
			(void)ircsprintf(pvar[*pcount], "%cb%s",
				what == MODE_ADD ? '+' : '-', tmpstr);
			(*pcount)++;				
			break;
		case MODE_EXCEPT:
			if (!param || *pcount >= MAXMODEPARAMS)
			{
				retval = 0;
				break;
			}
			retval = 1;
			tmpstr = pretty_mask(param);
			/* For bounce, we don't really need to worry whether
			 * or not it exists on our server.  We'll just always
			 * bounce it. */
			if (!bounce &&
			   ((what==MODE_ADD && add_exbanid(cptr, chptr, tmpstr))
			   || (what == MODE_DEL && del_exbanid(chptr, tmpstr))))
				break; /* already exists */
			(void)ircsprintf(pvar[*pcount], "%ce%s",
				what == MODE_ADD ? '+' : '-', tmpstr);
			(*pcount)++;				
			break;
		case MODE_LINK:
			if (IsULine(cptr, cptr)) {
				goto linkok;
			}
/*
#ifndef OPSLINK
			if (!(IsNetAdmin(cptr) || IsCoAdmin(cptr) || IsTechAdmin(cptr) && is_chanowner(cptr,chptr)))
			{
				sendto_one(cptr, ":%s NOTICE %s :*** Channel mode +L can only be set by the channel owner", me.name, cptr->name);				
				break;
			}
#endif
*/
			linkok:
			if (!chptr->mode.limit && what == MODE_ADD) {
				sendto_one(cptr, ":%s NOTICE %s :*** Channel mode +l <max> is requried for +L to be set", me.name, cptr->name);
				break;
			}
			if (!param || *pcount >= MAXMODEPARAMS)
			{
				retval = 0;
				break;
			}
			retval = 1;
			for (x = 0; x < *pcount; x++)
			{
				if (pvar[x][1] == 'L')
				{	/* don't allow user to change link
					 * more than once per command. */
					retval = 0;
					break;
				}
			}
			if (retval == 0) /* you can't break a case from loop */
				break;
			if (what == MODE_ADD)
			{
				if (!param || *pcount >= MAXMODEPARAMS)
				{
					retval = 0;
					break;
				}

				if (!stricmp(param, chptr->chname)) {
					if (MyClient(cptr))
						sendto_one(cptr, ":%s NOTICE %s :*** You can't link %s to itself", me.name, cptr->name, chptr->chname);
					break;
				}
				if (!IsChannelName(param)) {
					if (MyClient(cptr))
						sendto_one(cptr, err_str(ERR_NOSUCHCHANNEL), me.name, cptr->name, param);
					break;
				}
				if (!bounce) /* don't do the mode at all. */
					strncpyzt(chptr->mode.link, param,
						sizeof(chptr->mode.link));
				tmpstr = param;
			}
			else
			{
				if (!*chptr->mode.link)
					break; /* no change */
				strncpyzt(tmpbuf, chptr->mode.link,
					sizeof(tmpbuf));
				tmpstr = tmpbuf;
				if (!bounce)
					strcpy(chptr->mode.link, "");
			}
			retval = 1;
			
			(void)ircsprintf(pvar[*pcount], "%cL%s",
				what == MODE_ADD ? '+' : '-', tmpstr);
			(*pcount)++;				
			break;
		case MODE_FLOODLIMIT:
			if (!param || *pcount >= MAXMODEPARAMS)
			{
				retval = 0;
				break;
			}
			retval = 1;
			for (x = 0; x < *pcount; x++)
			{
				if (pvar[x][1] == 'f')
				{	/* don't allow user to change flood
					 * more than once per command. */
					retval = 0;
					break;
				}
			}
			if (retval == 0) /* you can't break a case from loop */
				break;
			if (what == MODE_ADD)
			{
				if (!bounce) /* don't do the mode at all. */
				{
					if (!param || *pcount >= MAXMODEPARAMS)
					{
						retval = 0;
						break;
					}

					/* like 1:1 and if its less than 3 chars then ahem.. */
					if (strlen(param) < 3)
					{
						sendto_one(cptr, ":%s NOTICE %s :*** Notice -- Flood Protection Syntax: /mode %s +f lines:seconds", me.name, cptr->name, chptr->chname);
						break;
					}
					/* may not contain other chars 
					   than 0123456789: & NULL */
					for (xp = param; *xp; xp++)
					{
						/* fast alpha check */
						if ( (*xp < '0') 
						  && (*xp > '9')
						  && (*xp != '\0')
						  && (*xp != ':')
						  && (*xp == '*' && xp != param))
						sendto_one(cptr, ":%s NOTICE %s :*** Notice -- Flood Protection Syntax: /mode %s +f lines:seconds", me.name, cptr->name, chptr->chname);
						  	break;
					}
					/* haven't got a : .. oh well */
					if (!strchr(param, ':'))
					{
						sendto_one(cptr, ":%s NOTICE %s :*** Notice -- Flood Protection Syntax: /mode %s +f lines:seconds", me.name, cptr->name, chptr->chname);
						break;
					}
					/* got multiple :'s .. omg */
					if (strchr(param, ':') != strrchr(param, ':'))
					{
						sendto_one(cptr, ":%s NOTICE %s :*** Notice -- Flood Protection Syntax: /mode %s +f lines:seconds", me.name, cptr->name, chptr->chname);
						break;
					}
					if (*param == '*' )
					{
						xzi = 1;
//					  	chptr->mode.kmode = 1;
					}
					 else
					{
						xzi = 0;
//					       chptr->mode.kmode = 0;
					}
					xp = index(param, ':');
					*xp = '\0';
					xxi = atoi((*param == '*' ? (param + 1) : param));
					xp++;
					xyi = atoi(xp);
					xp--;
					*xp = ':';
					if ((xxi == 0) || (xyi == 0))
						break;
					if ((chptr->mode.msgs == xxi) && (chptr->mode.per == xyi) && (chptr->mode.kmode == xzi))
						break;					chptr->mode.msgs = xxi;
					if ((xxi >= 101) || (xyi >= 101))
					{
						sendto_one(cptr, ":%s NOTICE %s :*** Notice -- Flood Protection Max 100 lines in 100 seconds. Decrease the settings and try again.", me.name, cptr->name, chptr->chname);
						break;
					}

					chptr->mode.msgs = xxi;
					chptr->mode.per = xyi;
					chptr->mode.kmode = xzi;
				}
				tmpstr = param;
				retval = 1;
			}
			else
			{
				if (!chptr->mode.msgs || !chptr->mode.per)
					break; /* no change */
				ircsprintf(tmpbuf, (chptr->mode.kmode > 0 ? "*%i:%i" : "%i:%i"), chptr->mode.msgs, chptr->mode.per);
				tmpstr = tmpbuf;
				if (!bounce)
				{
					chptr->mode.msgs = chptr->mode.per = chptr->mode.kmode =  0;
				}
				retval = 0;
			}
 			
			(void)ircsprintf(pvar[*pcount], "%cf%s",
				what == MODE_ADD ? '+' : '-', tmpstr);
			(*pcount)++;				
			break;
		default:
			
	}
	return(retval);	
}

/*
 * ListBits(bitvalue, bitlength);
 * written by Stskeeps
*/
char	*ListBits(long bits, long length) {
	char	*bitstr, *p;
	long	l,x,y;
        y = 1;
	bitstr = (char *) MyMalloc(length + 1);
        p = bitstr;
	for (l = 1; l <= length; l++) {
		if (bits & y)
			*p = '1';
		else
			*p = '0';
                p++;
		y = y + y;
	}
	*p = '\0';
        return (bitstr);
}


/* set_mode
 *	written by binary
 */
void set_mode(chptr, cptr, parc, parv, pcount, pvar, bounce)
aChannel	*chptr;
aClient	*cptr;
int 	parc;
u_int	*pcount;
char	bounce;
char	*parv[], pvar[MAXMODEPARAMS][MODEBUFLEN+3];
// long    opermode;
{
	Reg1 char *curchr;
	u_int	what = MODE_ADD;
	long	modetype;
	int paracount;
	char *tmpo;
	int *ip, xid;
	aCtab *tab = &cFlagTab[0];
	aCtab foundat;
	int	found = 0;
	paracount = 1;
	*pcount = 0;
	for (curchr = parv[0]; *curchr; curchr++)
	{
		switch(*curchr)
		{
			case '+':
				what = MODE_ADD;
				break;
			case '-':
				what = MODE_DEL;
				break;
			case '^':
				tmpo = (char *) ListBits(chptr->mode.mode, 64);
				sendto_one(cptr, ":%s NOTICE %s :*** %s mode is %li (0x%lx) [%s]",
				    me.name,
				    cptr->name,
				    chptr->chname,
				    chptr->mode.mode,
				    chptr->mode.mode,
				    tmpo);
				MyFree(tmpo);
				break;

			default:
				found = 0;
				tab = &cFlagTab[0];				while ((tab->mode != 0x0) && found == 0) {
					if (tab->flag == *curchr) {
						found = 1;
	       	    	 	                foundat = *tab;
					}
					tab++;
				}
				if (found == 1)
					modetype = foundat.mode;
				if (found == 0)
					modetype = 0;				
	
				if (modetype == 0)
				{
					if (!MyClient(cptr))
						break;
					    /* don't flood other servers */
					sendto_one(cptr,
						err_str(ERR_UNKNOWNMODE),
						me.name, cptr->name, *curchr);
					break;
				}
				if (parv[paracount] &&
						strlen(parv[paracount]) >=
						MODEBUFLEN)
					parv[paracount][MODEBUFLEN-1] = '\0';
				paracount += do_mode_char(chptr, modetype,
					*curchr, parv[paracount], what, cptr,
					pcount, pvar, bounce);
				break;
		}
	}
}

int DoesOp(modebuf)
char *modebuf;
{
  modebuf--; /* Is it possible that a mode starts with o and not +o ? */
  while (*++modebuf) if (*modebuf == 'h' || *modebuf=='o' || *modebuf=='v') return(1);
  return 0;
}

int sendmodeto_one(cptr, from, name, mode, param, creationtime)
Reg2 aClient *cptr;
char *from,*name,*mode,*param;
time_t creationtime;
{
	if ((IsServer(cptr) && DoesOp(mode) && creationtime) ||
 		IsULine(cptr,cptr))
	  sendto_one(cptr,":%s %s %s %s %s %lu", from,
	      (IsToken(cptr)?TOK_MODE:MSG_MODE), name, mode, param,
	      creationtime);
	else
	  sendto_one(cptr,":%s %s %s %s %s", from,
	      (IsToken(cptr)?TOK_MODE:MSG_MODE), name, mode, param);
}

char *pretty_mask(mask)
char *mask;
{ Reg1 char *cp;
  Reg2 char *user;
  Reg3 char *host;

  if ((user = index((cp = mask), '!'))) *user++ = '\0';
  if ((host = rindex(user ? user : cp, '@')))
  { *host++ = '\0';
    if (!user) return make_nick_user_host(NULL, cp, host); }
  else if (!user && index(cp, '.')) return make_nick_user_host(NULL, NULL, cp);
  return make_nick_user_host(cp, user, host);
}


/* Now let _invited_ people join thru bans, +i and +l.
 * Checking if an invite exist could be done only if a block exists,
 * but I'm not too fancy of the complicated structure that'd cause,
 * when optimization will hopefully take care of it. Most of the time
 * a user won't have invites on him anyway. -Donwulff
 */

static	int	can_join(cptr, sptr, chptr, key, link, parv)
aClient	*cptr, *sptr;
Reg2	aChannel *chptr;
char	*key, *link;
char	*parv[];
{
	Reg1	Link	*lp;
	int		ib = 1;
	/* 0 = noone 1 = IRCops 2 = Admins (override) */


/*    if ((chptr->mode.limit && chptr->users >= chptr->mode.limit) && *chptr->mode.link && !lp) {
            sendto_one(sptr, err_str(ERR_LINKCHANNEL), me.name, sptr->name,
                           chptr->chname, chptr->mode.link);
                    parv[0] = sptr->name;
			
			parv[1] = (chptr->mode.link);
			(void)channel_link(cptr, sptr, 2, parv);
			return (ERR_CHANNELISFULL);
	}
*/


/*if(join)count++; if(time>15)count--; if(count?<0)count=0; if(count>5)return(0);*/


    if ((chptr->mode.mode & MODE_OPERONLY) && IsAnOper(sptr)) {
    	goto admok;
    }
    if ((chptr->mode.mode & MODE_OPERONLY) && !IsAnOper(sptr)) {
        return (ERR_OPERONLY);
    }
    if ((chptr->mode.mode & MODE_ADMONLY)) {
			if (IsSkoAdmin(sptr))
	    	   		goto admok;
			return (ERR_ADMONLY);

    	}
		
admok:		
	if ((IsAnOper(sptr) && !(chptr->mode.mode & MODE_ADMONLY))) {
		return 0; /* may override */
	}
	if ((IsAnOper(sptr) && is_banned(cptr, sptr, chptr) && (chptr->mode.mode & MODE_OPERONLY)))
		return (ERR_BANNEDFROMCHAN); /* banned as an ircop at a +O cannot join */

/*	if (ib == 1)
		if (IsOper(sptr))
			return 0;
	else
		if (ib == 2)
			if (IsNetAdmin(sptr))
				return 0;
*/
	switch(ib) {
		case 1: if (IsOper(sptr)) return 0;
				break;
		case 2: if (IsSkoAdmin(sptr)) return 0;
				break;
		default: break;
	}	

	if (IsAnOper(sptr))
		return 0;

	if ((chptr->mode.limit && chptr->users >= chptr->mode.limit)) {
		if (chptr->mode.link) {
			if (*chptr->mode.link != '\0') {
				/* We are linked. */
				sendto_one(sptr, err_str(ERR_LINKCHANNEL), me.name, sptr->name,
				chptr->chname, chptr->mode.link);
				parv[0] = sptr->name;
				parv[1] = (chptr->mode.link);
				channel_link(cptr, sptr, 2, parv);						
				return -1;
			}
		}
		return (ERR_CHANNELISFULL);
	}



	if (*chptr->mode.key && (BadPtr(key) || mycmp(chptr->mode.key, key)))
		return (ERR_BADCHANNELKEY);

	for (lp = sptr->user->invited; lp; lp = lp->next)
		if (lp->value.chptr == chptr)
			break;

	if ((chptr->mode.mode & MODE_INVITEONLY) && !lp)
		return (ERR_INVITEONLYCHAN);

	if ((chptr->mode.limit && chptr->users >= chptr->mode.limit))

		return (ERR_CHANNELISFULL);

	if (is_banned(sptr, sptr, chptr) && !lp)
		return (ERR_BANNEDFROMCHAN);

	if ((chptr->mode.mode & MODE_RGSTRONLY) && !IsARegNick(sptr))
		return (ERR_NEEDREGGEDNICK);
		
	return 0;
}

/*
** Remove bells and commas from channel name
*/

void	clean_channelname(cn)
char	*cn;
{
	Reg1	u_char	*ch = (u_char *)cn;


	for (; *ch; ch++)
		/* Don't allow any control chars, the space, the comma,
		 * or the "non-breaking space" in channel names.
		 * Might later be changed to a system where the list of
		 * allowed/non-allowed chars for channels was a define
		 * or some such.
		 *   --Wizzu
		 */

		/*
		**
		** Dont allow special chars at all in the channel name.
		** Most of there are above 160 so.. - HAMLET
		**
		*/
		if (*ch < 33 || *ch == ',' || *ch >= 160)
		    {
			*ch = '\0';
			return;
		    }
}

/*
** Return -1 if mask is present and doesnt match our server name.
*/
static	int	check_channelmask(sptr, cptr, chname)
aClient	*sptr, *cptr;
char	*chname;
{
	Reg1	char	*s;

	if (*chname == '&')
		return -1;
	s = rindex(chname, ':');
	if (!s)
		return 0;

	s++;
	if (match(s, me.name) || (IsServer(cptr) && match(s, cptr->name)))
	    {
		if (MyClient(sptr))
			sendto_one(sptr, err_str(ERR_BADCHANMASK),
				   me.name, sptr->name, chname);
		return -1;
	    }
	return 0;
}

/*
**  Get Channel block for i (and allocate a new channel
**  block, if it didn't exists before).
*/
static	aChannel *get_channel(cptr, chname, flag)
aClient *cptr;
char	*chname;
int	flag;
    {
	Reg1	aChannel *chptr;
	int	len;

	if (BadPtr(chname))
		return NULL;

	len = strlen(chname);
	if (MyClient(cptr) && len > CHANNELLEN)
	    {
		len = CHANNELLEN;
		*(chname+CHANNELLEN) = '\0';
	    }
	if ((chptr = find_channel(chname, (aChannel *)NULL)))
		return (chptr);
	if (flag == CREATE)
	    {
		chptr = (aChannel *)MyMalloc(sizeof(aChannel) + len);
		bzero((char *)chptr, sizeof(aChannel));
		strncpyzt(chptr->chname, chname, len+1);
		if (channel)
			channel->prevch = chptr;
		chptr->prevch = NULL;
		chptr->nextch = channel;
		chptr->creationtime = MyClient(cptr)?TStime():(time_t)0;
		channel = chptr;
		(void)add_to_channel_hash_table(chname, chptr);
	    }
	return chptr;
    }

/*
 * Slight changes in routine, now working somewhat symmetrical:
 *   First try to remove the client & channel pair to avoid duplicates
 *   Second check client & channel invite-list lengths and remove tail
 *   Finally add new invite-links to both client and channel
 * Should U-lined clients have higher limits?   -Donwulff
 */

static	void	add_invite(cptr, chptr)
aClient *cptr;
aChannel *chptr;
{
	Reg1	Link	*inv, *tmp;

	del_invite(cptr, chptr);
	/*
	 * delete last link in chain if the list is max length
	 */
	if (list_length(cptr->user->invited) >= MAXCHANNELSPERUSER)
	    {
/*		This forgets the channel side of invitation     -Vesa
		inv = cptr->user->invited;
		cptr->user->invited = inv->next;
		free_link(inv);
*/
		for (tmp = cptr->user->invited; tmp->next; tmp = tmp->next)
			;
		del_invite(cptr, tmp->value.chptr);
 
	    }
	/* We get pissy over too many invites per channel as well now,
	 * since otherwise mass-inviters could take up some major
	 * resources -Donwulff
	 */
	if (list_length(chptr->invites) >= MAXCHANNELSPERUSER) {
		for (tmp = chptr->invites; tmp->next; tmp = tmp->next)
			;
		del_invite(tmp->value.cptr, chptr);
	}
	/*
	 * add client to the beginning of the channel invite list
	 */
	inv = make_link();
	inv->value.cptr = cptr;
	inv->next = chptr->invites;
	chptr->invites = inv;
	/*
	 * add channel to the beginning of the client invite list
	 */
	inv = make_link();
	inv->value.chptr = chptr;
	inv->next = cptr->user->invited;
	cptr->user->invited = inv;
}

/*
 * Delete Invite block from channel invite list and client invite list
 */
void	del_invite(cptr, chptr)
aClient *cptr;
aChannel *chptr;
{
	Reg1	Link	**inv, *tmp;

	for (inv = &(chptr->invites); (tmp = *inv); inv = &tmp->next)
		if (tmp->value.cptr == cptr)
		    {
			*inv = tmp->next;
			free_link(tmp);
			break;
		    }

	for (inv = &(cptr->user->invited); (tmp = *inv); inv = &tmp->next)
		if (tmp->value.chptr == chptr)
		    {
			*inv = tmp->next;
			free_link(tmp);
			break;
		    }
}

/*
**  Subtract one user from channel i (and free channel
**  block, if channel became empty).
*/
static	void	sub1_from_channel(chptr)
Reg1	aChannel *chptr;
{
	Ban	*ban;
	Link	*lp;

	if (--chptr->users <= 0)
	    {
		/*
		 * Now, find all invite links from channel structure
		 */
		while ((lp = chptr->invites))
			del_invite(lp->value.cptr, chptr);

		while (chptr->banlist)
		{
			ban = chptr->banlist;
			chptr->banlist = ban->next;
			MyFree(ban->banstr);
			MyFree(ban->who);
			free_ban(ban);
		}
		while (chptr->exlist)
		{
			ban = chptr->exlist;
			chptr->exlist = ban->next;
			MyFree(ban->banstr);
			MyFree(ban->who);
			free_ban(ban);
		}
		if (chptr->prevch)
			chptr->prevch->nextch = chptr->nextch;
		else
			channel = chptr->nextch;
		if (chptr->nextch)
			chptr->nextch->prevch = chptr->prevch;
		(void)del_from_channel_hash_table(chptr->chname, chptr);
		MyFree((char *)chptr);
	    }
}



/*
 * Channel Link
 */
int     channel_link(cptr, sptr, parc, parv)
Reg2    aClient *cptr, *sptr;
int     parc;
char    *parv[];
{
        static  char    jbuf[BUFSIZE], *tmp;
        Reg1    Link    *lp;
        Reg3    aChannel *chptr;
        Reg4    char    *name, *key = NULL, *link = NULL;
        int     i,i1, flags = 0, zombie = 0;
        char    *p = NULL, *p2 = NULL;

        if (check_registered_user(sptr))
                return 0;


        if (parc < 2 || *parv[1] == '\0')
            {
                sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS),
                           me.name, parv[0], "JOIN");
                return 0;
            }

        *jbuf = '\0';
        /*
        ** Rebuild list of channels joined to be the actual result of the
        ** JOIN.  Note that "JOIN 0" is the destructive problem.
        */
		bouncedtimes++;
        if (bouncedtimes > MAXBOUNCE) {
        	/* bounced too many times */
        	sendto_one(sptr, ":%s NOTICE %s :*** Couldn't join %s ! - Link setting was too bouncy",
        		me.name,
        		sptr->name,
        		parv[1]);
        	return;
        }
        for (i = 0, name = strtoken(&p, parv[1], ","); name;
             name = strtoken(&p, NULL, ","))
            {
                /* pathological case only on longest channel name.
                ** If not dealt with here, causes desynced channel ops
                ** since ChannelExists() doesn't see the same channel
                ** as one being joined. cute bug. Oct 11 1997, Dianora/comstud
                ** Copied from Dianora's "hybrid 5" ircd.
                */

                if(strlen(name) >  CHANNELLEN)  /* same thing is done in get_channel() */
                        name[CHANNELLEN] = '\0';

                if (MyConnect(sptr))
                        clean_channelname(name);
                if (check_channelmask(sptr, cptr, name)==-1)
                        continue;
                if (*name == '&')
                        continue;
                if (*name == '0' && !atoi(name))
                    {
                        (void)strcpy(jbuf, "0");
                        i = 1;
                        continue;
                    }
                else if (!IsChannelName(name))
                    {
                        if (MyClient(sptr))
                                sendto_one(sptr, err_str(ERR_NOSUCHCHANNEL),
                                           me.name, parv[0], name);
                        continue;
                    }
                if (*jbuf)
                        (void)strcat(jbuf, ",");
                (void)strncat(jbuf, name, sizeof(jbuf) - i - 1);
                i += strlen(name)+1;
            }
        (void)strcpy(parv[1], jbuf);

        p = NULL;
        if (parv[2])
        if (parv[2])
                key = strtoken(&p2, parv[2], ",");
        parv[2] = NULL; /* for m_names call later, parv[parc] must == NULL */
        for (name = strtoken(&p, jbuf, ","); name;
             key = (key) ? strtoken(&p2, NULL, ",") : NULL,
             name = strtoken(&p, NULL, ","))
            {
                /*
                ** JOIN 0 sends out a part for all channels a user
                ** has joined.
                */
                if (*name == '0' && !atoi(name))
                    {
                        while ((lp = sptr->user->channel))
                            {
                                chptr = lp->value.chptr;
                                  sendto_channel_butserv(chptr, sptr, PartFmt,
                                      parv[0], chptr->chname);
                                remove_user_from_channel(sptr, chptr);
                            }
                        sendto_serv_butone(cptr, ":%s JOIN 0", parv[0]);
                        continue;
                    }

                if (MyConnect(sptr))
                    {
                        /*
                        ** local client is first to enter previously nonexistant
                        ** channel so make them (rightfully) the Channel
                        ** Operator.
                        */
                        if (!IsModelessChannel(name))
                            flags = (ChannelExists(name)) ? CHFL_DEOPPED :
                                                            CHFL_CHANOP;
                        else
                            flags = CHFL_DEOPPED;

                        if (sptr->user->joined >= MAXCHANNELSPERUSER)
                            {
                                sendto_one(sptr, err_str(ERR_TOOMANYCHANNELS),
                                           me.name, parv[0], name);
                                return 0;
                            }
                    }

                chptr = get_channel(sptr, name, CREATE);
                if (chptr && (lp=find_user_link(chptr->members, sptr)))
                  continue;

                	if (!MyConnect(sptr)) flags = CHFL_DEOPPED;
                 	 if (sptr->flags & FLAGS_TS8) flags|=CHFL_SERVOPOK;

		i1 = 0;
		if (chptr == NULL)
			return;

                if (!chptr ||
                    (MyConnect(sptr) && (i = can_join(cptr, sptr, chptr, key, link, parv))))
                    {
                        if ( i != -1 ) 
	                {        sendto_one(sptr, err_str(i),
        	                           me.name, parv[0], name);
                        }
                        continue;
                    }

                /*
                **  Complete user entry to the new channel (if any)
                */
                add_user_to_channel(chptr, sptr, flags);
                /*
                ** notify all other users on the new channel
                */
				if (!IsHiding(sptr))
	                sendto_channel_butserv(chptr, sptr, ":%s JOIN :%s",
                                        parv[0], name);
    			else {
    				if (MyClient(sptr)) 
    					sendto_one(sptr, ":%s!%s@%s JOIN :%s", sptr->name, sptr->user->username,
							(IsHidden(sptr) ? sptr->user->virthost : sptr->user->realhost),
							 name);
    				sendto_umode(UMODE_NETADMIN|UMODE_TECHADMIN, "*** Invisible(+I) user %s joined %s", sptr->name, chptr->chname);
    			}
                sendto_match_servs(chptr, cptr, ":%s JOIN :%s", parv[0], name);

                if (MyClient(sptr))
                    {
                        /*
                        ** Make a (temporal) creationtime, if someone joins
                        ** during a net.reconnect : between remote join and
                        ** the mode with TS. --Run
                        */
                        if (chptr->creationtime == 0)
                        { chptr->creationtime = TStime();
                          sendto_match_servs(chptr, cptr, ":%s MODE %s + %lu",
                              me.name, name, chptr->creationtime);
                        }
                        del_invite(sptr, chptr);
                        if (flags & CHFL_CHANOP)
                                sendto_match_servs(chptr, cptr,
                                  ":%s MODE %s +o %s %lu",
                                  me.name, name, parv[0], chptr->creationtime);
                        if (chptr->topic[0] != '\0') {
                                sendto_one(sptr, rpl_str(RPL_TOPIC), me.name,
                                           parv[0], name, chptr->topic);
                                sendto_one(sptr, rpl_str(RPL_TOPICWHOTIME),
                                            me.name, parv[0], name,
                                            chptr->topic_nick,
                                            chptr->topic_time);
                              }
                        parv[1] = name;
                        (void)m_names(cptr, sptr, 2, parv);
                    	bouncedtimes = 0;
                    }

            }
        return 0;
}


/*
** m_join
**	parv[0] = sender prefix
**	parv[1] = channel
**	parv[2] = channel password (key)
*/
int	m_join(cptr, sptr, parc, parv)
Reg2	aClient *cptr, *sptr;
int	parc;
char	*parv[];
{
	static	char	jbuf[BUFSIZE], *tmp;
	Reg1	Link	*lp;
	Reg3	aChannel *chptr;
	Reg4	char	*name, *key = NULL, *link = NULL;
	int	i, flags = 0, zombie = 0;
	char	*p = NULL, *p2 = NULL;

	if (check_registered_user(sptr))
		return 0;

	bouncedtimes = 0;

	if (parc < 2 || *parv[1] == '\0')
	    {
		sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS),
			   me.name, parv[0], "JOIN");
		return 0;
	    }
	*jbuf = '\0';
	/*
	** Rebuild list of channels joined to be the actual result of the
	** JOIN.  Note that "JOIN 0" is the destructive problem.
	*/
	for (i = 0, name = strtoken(&p, parv[1], ","); name;
	     name = strtoken(&p, NULL, ","))
	    {
		/* pathological case only on longest channel name.
		** If not dealt with here, causes desynced channel ops
		** since ChannelExists() doesn't see the same channel
		** as one being joined. cute bug. Oct 11 1997, Dianora/comstud
		** Copied from Dianora's "hybrid 5" ircd.
		*/

		if(strlen(name) >  CHANNELLEN)   /* same thing is done in get_channel() */
			name[CHANNELLEN] = '\0';
		
		if (MyConnect(sptr))
			clean_channelname(name);
		if (check_channelmask(sptr, cptr, name)==-1) 
			continue;
		if (*name == '&' && !MyConnect(sptr))
			continue;
	

		/*
		**
		** This looks like a leftover from join 0 to me - HAMLET
		**
		*/
/*
		if (*name == '0' && !atoi(name))
		    {
		    	(void)strcpy(jbuf, "0");
		        i = 1;
		        continue;
		    }
*/
		else if (!IsChannelName(name))
		    {
			if (MyClient(sptr))
				sendto_one(sptr, err_str(ERR_NOSUCHCHANNEL),
					   me.name, parv[0], name);
			continue;
		    }
		if (*jbuf)
			(void)strcat(jbuf, ",");
		(void)strncat(jbuf, name, sizeof(jbuf) - i - 1);
		i += strlen(name)+1;
	    }
	(void)strcpy(parv[1], jbuf);

	p = NULL;
	if (parv[2])
		key = strtoken(&p2, parv[2], ",");
	parv[2] = NULL;	/* for m_names call later, parv[parc] must == NULL */
	for (name = strtoken(&p, jbuf, ","); name;
	     key = (key) ? strtoken(&p2, NULL, ",") : NULL,
	     name = strtoken(&p, NULL, ","))
	    {
		if (MyConnect(sptr))
		    {
			/*
			** local client is first to enter previously nonexistant
			** channel so make them (rightfully) the Channel
			** Operator.
			*/
			if (!IsModelessChannel(name))
			    flags = (ChannelExists(name)) ? CHFL_DEOPPED :
							    CHFL_CHANOP;
			else
			    flags = CHFL_DEOPPED;

			if (!IsAnOper(sptr)) /* Opers can join unlimited channels */
			{
				if (sptr->user->joined >= MAXCHANNELSPERUSER)
			    	{
					sendto_one(sptr, err_str(ERR_TOOMANYCHANNELS),
					   	me.name, parv[0], name);
					return 0;
			    	}
			}
			/*
			**
			** This looks like a good place to put in the initial system
			** for restricting channels - HAMLET
			**
			*/

			if (crlines)
			{
				if (channel_canjoin(sptr, name) != 1)
				{
					if (cannotjoin_msg)
						sendto_one(sptr, ":%s NOTICE %s :*** Notice -- Cannot join %s. %s", me.name, sptr->name, name, cannotjoin_msg);
						sendto_one(sptr, ":%s NOTICE %s :*** Notice -- Please do a \2/SETTINGS CHANNELS\2 to see a list of all channels that can be joined.", me.name, sptr->name);
					return 0;
				}
			}
		    }

		chptr = get_channel(sptr, name, CREATE);

/*
**
** Need to look into the following piece of code and test - HAMLET
**
*/
/*

  if (chptr) {
sendto_one (sptr, err_str (ERR_NOSUCHCHANNEL),
    me.name, parv[0], name);
return(0);
  }
*/


                if (chptr && (lp=find_user_link(chptr->members, sptr)))
		  continue; 
		
		if (!MyConnect(sptr)) flags = CHFL_DEOPPED;
		if (sptr->flags & FLAGS_TS8) flags|=CHFL_SERVOPOK;

		if (!chptr ||
		    (MyConnect(sptr) && (i = can_join(cptr, sptr, chptr, key, link, parv))))
		    { 
				
				if (i != -1)
					sendto_one(sptr, err_str(i),
						   me.name, parv[0], name);
				
				/* uhm? was *chptr ??? *NULL = dangerous */


			continue;
				
   			}
   		if (is_banned(cptr, sptr, chptr) && IsOper(sptr))
   		{
   			sendto_umode(UMODE_NGLOBAL, "*** Banwalk [IRCop: %s] [Channel: %s]", sptr->name, chptr->chname);
		    }

		/*
		**  Complete user entry to the new channel (if any)
		*/
		add_user_to_channel(chptr, sptr, flags);
		/*
		** notify all other users on the new channel
		*/
		if (!IsHiding(sptr))
			sendto_channel_butserv(chptr, sptr, ":%s JOIN :%s",
						parv[0], chptr->chname);
		else {
				if (MyClient(sptr))
					sendto_one(sptr, ":%s!%s@%s JOIN :%s", sptr->name, sptr->user->username,
					(IsHidden(sptr) ? sptr->user->virthost : sptr->user->realhost),
					 chptr->chname);
				if (MyClient(sptr)) {
					sendto_umode(UMODE_ADMIN, "*** [+I] %s invisible joined %s", sptr->name, chptr->chname); 
//					sendto_serv_butone(&me, ":%s SMO A :[+I] %s invisible joined %s", me.name, sptr->name, chptr->chname);
				}
			}

		sendto_match_servs(chptr, cptr, ":%s JOIN :%s", parv[0], chptr->chname);

		if (MyClient(sptr))
		    {
		        /*
			** Make a (temporal) creationtime, if someone joins
			** during a net.reconnect : between remote join and
			** the mode with TS. --Run
			*/
		        if (chptr->creationtime == 0)
		        { chptr->creationtime = TStime();
			  sendto_match_servs(chptr, cptr, ":%s MODE %s + %lu",
			      me.name, chptr->chname, chptr->creationtime);
			}
			del_invite(sptr, chptr);
			if (flags & CHFL_CHANOP)
				sendto_match_servs(chptr, cptr,
				  ":%s MODE %s +o %s %lu",
				  me.name, chptr->chname, parv[0], chptr->creationtime);
			if (chptr->topic[0] != '\0') {
				sendto_one(sptr, rpl_str(RPL_TOPIC), me.name,
					   parv[0], name, chptr->topic);
				sendto_one(sptr, rpl_str(RPL_TOPICWHOTIME),
					    me.name, parv[0], name,
					    chptr->topic_nick,
					    chptr->topic_time);
			      }
			parv[1] = chptr->chname;
			(void)m_names(cptr, sptr, 2, parv);
		    }
	    }
	bouncedtimes = 0;
	return 0;
}

/*
** m_part
**	parv[0] = sender prefix
**	parv[1] = channel
**	parv[2] = comment (added by Lefler)
*/
int	m_part(cptr, sptr, parc, parv)
aClient *cptr, *sptr;
int	parc;
char	*parv[];
    {
    int		xx = 0;
	Reg1	aChannel *chptr;
	Reg2  Link	*lp;
	char	*p = NULL, *name;
	register char *comment = (parc > 2 && parv[2]) ? parv[2] : NULL;

	if (check_registered_user(sptr))
		return 0;

        sptr->flags&=~FLAGS_TS8;

	if (parc < 2 || parv[1][0] == '\0')
	    {
		sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS),
			   me.name, parv[0], "PART");
		return 0;
	    }

#ifdef	V28PlusOnly
	*buf = '\0';
#endif

	for (; (name = strtoken(&p, parv[1], ",")); parv[1] = NULL)
	    {
		chptr = get_channel(sptr, name, 0);
		if (!chptr)
		    {
			sendto_one(sptr, err_str(ERR_NOSUCHCHANNEL),
				   me.name, parv[0], name);
			continue;
		    }
		if (check_channelmask(sptr, cptr, name))
			continue;
		/* Do not use IsMember here: zombies must be able to part too */
		if (!(lp=find_user_link(chptr->members, sptr)))
		    {
			/* Normal to get get when our client did a kick
			** for a remote client (who sends back a PART),
			** so check for remote client or not --Run
			*/
			if (MyClient(sptr))
			  sendto_one(sptr, err_str(ERR_NOTONCHANNEL),
		    	      me.name, parv[0], name);
			continue;
		    }
		/*
		**  Remove user from the old channel (if any)
		*/
#ifdef	V28PlusOnly
		if (*buf)
			(void)strcat(buf, ",");
		(void)strcat(buf, name);
#else
                if (parc < 3)
		  sendto_match_servs(chptr, cptr, PartFmt, parv[0], chptr->chname);
		else
		  sendto_match_servs(chptr, cptr, PartFmt2, parv[0], chptr->chname, comment);
#endif
		if (1)
		{
			if (!IsHiding(sptr))
			{
	 		       	if (parc < 3)
	 		       	{
					sendto_channel_butserv(chptr, sptr, PartFmt, parv[0], chptr->chname);
				} 
			  		else
			  	{
			    			sendto_channel_butserv(chptr, sptr, PartFmt2, parv[0], chptr->chname, comment);
			    	}
        		}
 				else
 			{
 				if (MyClient(sptr))
 				{
 					sendto_umode(UMODE_ADMIN, "*** [+I] %s invisible parted %s", sptr->name, chptr->chname);
// 					sendto_serv_butone(&me, ":%s SMO A :[+I] %s invisible parted %s", me.name, sptr->name, chptr->chname);
 				}
				if (MyClient(sptr))
					/* awful hack .. */
					 if (parc < 3)
					 	sendto_one(sptr, ":%s!%s@%s PART %s", sptr->name, sptr->user->username,
						(IsHidden(sptr) ? sptr->user->virthost : sptr->user->realhost),
							 chptr->chname);
						else
						sendto_one(sptr, ":%s!%s@%s PART %s :%s", sptr->name, sptr->user->username,
							(IsHidden(sptr) ? sptr->user->virthost : sptr->user->realhost),
							 chptr->chname, comment);
		       	}
		 	remove_user_from_channel(sptr, chptr);
		 }
	}

#ifdef	V28PlusOnly
//	if (*buf)
		sendto_serv_butone(cptr, PartFmt, parv[0], buf);
#endif
	return 0;
    }

/*
** m_kick
**	parv[0] = sender prefix
**	parv[1] = channel
**	parv[2] = client to kick
**	parv[3] = kick comment
*/
int	m_kick(cptr, sptr, parc, parv)
aClient *cptr, *sptr;
int	parc;
char	*parv[];
{
	aClient *who;
	aChannel *chptr;
	int	chasing = 0;
	char	*comment, *name, *p = NULL, *user, *p2 = NULL;
	Link *lp,*lp2;

	if (check_registered(sptr))
		return 0;

	sptr->flags&=~FLAGS_TS8;

	if (parc < 3 || *parv[1] == '\0')
	    {
		sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS),
			   me.name, parv[0], "KICK");
		return 0;
	    }

	comment = (BadPtr(parv[3])) ? parv[0] : parv[3];
	if (strlen(comment) > (size_t) TOPICLEN)
		comment[TOPICLEN] = '\0';

	*nickbuf = *buf = '\0';

	for (; (name = strtoken(&p, parv[1], ",")); parv[1] = NULL)
	    {
		chptr = get_channel(sptr, name, !CREATE);
		if (!chptr)
		    {
			sendto_one(sptr, err_str(ERR_NOSUCHCHANNEL),
				   me.name, parv[0], name);
			continue;
		    }
		if (check_channelmask(sptr, cptr, name))
			continue;
		if (!IsServer(cptr) && !IsAdmin(sptr) && !IsULine(cptr,sptr) && !is_chan_op(sptr, chptr) && !is_halfop(sptr,chptr))
		    {
			sendto_one(sptr, err_str(ERR_CHANOPRIVSNEEDED),
				   me.name, parv[0], chptr->chname);
			continue;
		    }

	        lp2=find_user_link(chptr->members, sptr);
		for (; (user = strtoken(&p2, parv[2], ",")); parv[2] = NULL)
		    {   
			if (!(who = find_chasing(sptr, user, &chasing)))
				continue; /* No such user left! */


/*
**
** If the user is trying to kick a hidden user
** Tell him that the user aint present - HAMLET
**
*/
		if (!(IsServer(sptr) || IsULine(cptr,sptr) || IsNetAdmin(sptr) || IsTechAdmin(sptr))) {
		if ((who->umodes & UMODE_HIDING) && MyClient(sptr)) {

                          sendto_one(sptr, err_str(ERR_NOSUCHNICK),
                                           me.name, parv[0], user, name);
			 sendto_one(who,":%s Notice %s :%s tried to kick you from %s",
				me.name, who->name, parv[0], parv[1]);
			  return 0;
		   }
		}


		/* if the user is +p, prevent a kick from local user */
		if (!IsServer(cptr) || !IsULine(cptr,sptr) || !IsSuper(cptr)) {
		if ((who->umodes & UMODE_SUPER) && MyClient(sptr)) {
			 sendto_one(sptr,":%s Notice %s :%s cannot be kicked from %s (Protected IRC Operator)",
				me.name, parv[0], parv[2], parv[1]);
			sendto_netglobal("Overriding Kick for: %s on channel: %s made by: %s (%s@%s)",
				parv[2], parv[1], parv[0], cptr->user->username, cptr->user->host);
			sendto_serv_butone(&me, ":%s NETG :Overriding Kick for: %s on channel: %s made by: %s (%s@%s)",
				me.name, parv[2], parv[1], parv[0], cptr->user->username, cptr->user->host);
			  return 0;

			}
			}
			  if (is_chan_op(who,chptr) && is_halfop(sptr, chptr) && !is_chan_op(sptr, chptr) && !IsULine(cptr, sptr)) {
					sendto_one(sptr, ":%s NOTICE %s :*** You cannot kick channel operators on %s if you only are halfop", me.name, sptr->name, chptr->chname);
					return 0;							  
			  }


                        if ((lp=find_user_link(chptr->members, who)) &&
			    (!(lp->flags & CHFL_ZOMBIE) || IsServer(sptr)))


	                {
                          if (IsServices(who) && !IsULine(cptr, sptr) &&
                                who != sptr)
                            {
                              if (IsServer(cptr))
                                {
                                  sendto_one(cptr, ":%s JOIN :%s",
                                    who->name, chptr->chname);
                                  if (lp->flags & CHFL_CHANOP)
                                    sendto_one(cptr, ":%s MODE %s +o %s",
                                      me.name, chptr->chname, who->name);
                                  if (lp->flags & CHFL_VOICE)
                                    sendto_one(cptr, ":%s MODE %s +v %s",
                                      me.name, chptr->chname, who->name);
                                }
                     sendto_one(sptr, ":%s NOTICE %s :*** [%s] Cannot kill, deop, or kick Services Clients.", me.name, parv[0], user);
                          continue;

                            }

                            if (lp)
                              sendto_channel_butserv(chptr, sptr,
                                  ":%s KICK %s %s :%s", parv[0],
                                  name, who->name, comment);
                            sendto_match_servs(chptr, cptr,
                                               ":%s KICK %s %s :%s",
                                               parv[0], name,
                                               who->name, comment);
                            if (lp)
                            {
                              remove_user_from_channel(who, chptr);

                            }
                        }
                        else if (MyClient(sptr))
                          sendto_one(sptr, err_str(ERR_USERNOTINCHANNEL),
                                           me.name, parv[0], user, name);
                        if (!IsServer(cptr) || !IsULine(cptr,sptr))
                          break;
                } /* loop on parv[2] */
                if (!IsServer(cptr) || !IsULine(cptr,sptr))
                  break;
            } /* loop on parv[1] */

        return 0;
}


int	count_channels(sptr)
aClient	*sptr;
{
Reg1	aChannel	*chptr;
	Reg2	int	count = 0;

	for (chptr = channel; chptr; chptr = chptr->nextch)
		if (SecretChannel(chptr))
		    {
			if (IsAnOper(sptr))
				count++;
		    }
		else

			count++;
	return (count);
}

/*
** m_topic
**	parv[0] = sender prefix
**	parv[1] = topic text
**
**	For servers using TS: (Lefler)
**	parv[0] = sender prefix
**	parv[1] = channel list
**	parv[2] = topic nickname
**	parv[3] = topic time
**	parv[4] = topic text
*/
int	m_topic(cptr, sptr, parc, parv)
aClient *cptr, *sptr;
int	parc;
char	*parv[];
    {
	aChannel *chptr = NullChn;
	char	*topic = NULL, *name, *p = NULL, *tnick = NULL;
	time_t	ttime = 0;
	
	if (check_registered(sptr))
		return 0;

	if (parc < 2)
	    {
		sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS),
			   me.name, parv[0], "TOPIC");
		return 0;
	    }

	for (; (name = strtoken(&p, parv[1], ",")); parv[1] = NULL)
	    {
		if (parc > 1 && IsChannelName(name))
		    {
			chptr = find_channel(name, NullChn);
			if (!chptr || (!IsMember(sptr, chptr) &&
			    !IsServer(sptr) && !IsULine(cptr,sptr)))
			    {
				sendto_one(sptr, err_str(ERR_NOTONCHANNEL),
					   me.name, parv[0], name);
				continue;
			    }				
			if (parc > 2)
				topic = parv[2];
			if (parc > 4) {
				tnick = parv[2];
				ttime = atol(parv[3]);
				topic = parv[4];
			/*
			**
			** Lets make sure U Lined clients get the current time
			** when submitting the time as 0. - HAMLET
			**
			*/
			if (IsULine(sptr, sptr) && (ttime == 0))
			{
				ttime = TStime();
			}
			}
		    }

		if (!chptr)
		    {
			sendto_one(sptr, rpl_str(RPL_NOTOPIC),
				   me.name, parv[0], name);
			return 0;
		    }

		if (check_channelmask(sptr, cptr, name))
			continue;
	
		if (!topic)  /* only asking  for topic  */
		    {
			if (chptr->topic[0] == '\0')
			sendto_one(sptr, rpl_str(RPL_NOTOPIC),
				   me.name, parv[0], chptr->chname);
			else {
				sendto_one(sptr, rpl_str(RPL_TOPIC),
					   me.name, parv[0],
					   chptr->chname, chptr->topic);
				sendto_one(sptr, rpl_str(RPL_TOPICWHOTIME),
					   me.name, parv[0], chptr->chname,
					   chptr->topic_nick,
					   chptr->topic_time);
		    } 
		    } 
		else if (ttime && topic && (IsServer(sptr) || IsULine(cptr,sptr)))
		    {
//			if (!chptr->topic_time || ttime < chptr->topic_time)
			if (IsULine(sptr, sptr) || !chptr->topic_time ||
				(ttime <= chptr->topic_time))
			   {
				/*
				**
				** Let U-Lined clients override the topic always. - HAMLET
				**
				*/

				if (StrEq(chptr->topic, topic) && !IsULine(sptr, sptr))
					continue;
				/* setting a topic */
				strncpyzt(chptr->topic, topic, sizeof(chptr->topic));
				strcpy(chptr->topic_nick, tnick);
				chptr->topic_time = ttime;
				sendto_match_servs(chptr, cptr,":%s TOPIC %s %s %lu :%s",
					   parv[0], chptr->chname,
					   chptr->topic_nick, chptr->topic_time,
					   chptr->topic);
				sendto_channel_butserv(chptr, sptr, ":%s TOPIC %s :%s (%s)",
					       parv[0], chptr->chname,
					       chptr->topic, chptr->topic_nick);
			   }
		    }
		else if (((chptr->mode.mode & MODE_TOPICLIMIT) == 0 ||
			  is_chan_op(sptr, chptr)) || IsSkoAdmin(sptr) || IsULine(cptr,sptr)
			  || is_halfop(sptr, chptr)  && topic)
                    {
			/* setting a topic */
/*
			if (IsOper(sptr) && !(is_halfop(sptr, chptr) || IsULine(cptr,sptr) || is_chan_op(sptr, chptr)) && (chptr->mode.mode & MODE_TOPICLIMIT))
			{
				sendto_umode(UMODE_NGLOBAL, "*** OperTopic [IRCop: %s] - [Channel: %s] - [Topic: %s]", sptr->name, chptr->chname, topic);
			}
*/
			strncpyzt(chptr->topic, topic, sizeof(chptr->topic));
			strcpy(chptr->topic_nick, sptr->name);
                        if (ttime && IsServer(cptr))
			  chptr->topic_time = ttime;
			else
			  chptr->topic_time = TStime();
			sendto_match_servs(chptr, cptr,":%s TOPIC %s %s %lu :%s",
					   parv[0], chptr->chname, parv[0],
					   chptr->topic_time, chptr->topic);
			sendto_channel_butserv(chptr, sptr, ":%s TOPIC %s :%s",
					       parv[0],
					       chptr->chname, chptr->topic);
		    }
		else
		      sendto_one(sptr, err_str(ERR_CHANOPRIVSNEEDED),
				 me.name, parv[0], chptr->chname);
	    }
	return 0;
    }

/*
** m_invite
**	parv[0] - sender prefix
**	parv[1] - user to invite
**	parv[2] - channel number
*/
int	m_invite(cptr, sptr, parc, parv)
aClient *cptr, *sptr;
int	parc;
char	*parv[];
    {
	aClient *acptr;
	aChannel *chptr;

	if (check_registered_user(sptr))
		return 0;

	if (parc < 3 || *parv[1] == '\0')
	    {
		sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS),
			   me.name, parv[0], "INVITE");
		return -1;
	    }

/* Prevent users from inviting to a channel they are not on -Studded */
	if (!(ChannelExists(parv[2])))
           {     sendto_one(sptr, err_str(ERR_NOTONCHANNEL),
                           me.name, parv[0], parv[2]);
                return -1;
	   }


	if (!(acptr = find_person(parv[1], (aClient *)NULL)))
	    {
		sendto_one(sptr, err_str(ERR_NOSUCHNICK),
			   me.name, parv[0], parv[1]);
		return -1;
	    }
	if (MyConnect(sptr))
		clean_channelname(parv[2]);
	if (check_channelmask(sptr, cptr, parv[2]))
		return 0;
	if (!(chptr = find_channel(parv[2], NullChn)))
	    {

		sendto_prefix_one(acptr, sptr, ":%s INVITE %s :%s",
				  parv[0], parv[1], parv[2]);
		return 0;
	    }

/*
**
** We want IRCOp's to be able to invite past this also.
** This should be in the networks settings file eventually - HAMLET
**
*/

	if (chptr->mode.mode & MODE_NOINVITE)
		if (!IsULine(cptr,sptr) || !IsAnOper(sptr)) {
			sendto_one(sptr, err_str(ERR_NOINVITE),
				me.name, parv[0], parv[2]);
			return -1;
		} 

	if (chptr && !IsMember(sptr, chptr) && !IsULine(cptr,sptr))
	    {
		sendto_one(sptr, err_str(ERR_NOTONCHANNEL),
			   me.name, parv[0], parv[2]);
		return -1;
	    }

	if (IsMember(acptr, chptr))
	    {
		sendto_one(sptr, err_str(ERR_USERONCHANNEL),
			   me.name, parv[0], parv[1], parv[2]);
		return 0;
	    }
	if (chptr && (chptr->mode.mode & MODE_INVITEONLY))
	    {
		if (!is_chan_op(sptr, chptr) && !IsULine(cptr,sptr))
		    {
			sendto_one(sptr, err_str(ERR_CHANOPRIVSNEEDED),
				   me.name, parv[0], chptr->chname);
			return -1;
		    }
		else if (!IsMember(sptr, chptr) && !IsULine(cptr,sptr))
		    {
			sendto_one(sptr, err_str(ERR_CHANOPRIVSNEEDED),
				   me.name, parv[0],
				   ((chptr) ? (chptr->chname) : parv[2]));
			return -1;
		    }
	    }

	if (MyConnect(sptr))
	    {
		if(check_for_target_limit(sptr, acptr, acptr->name))
			return 0;

		sendto_one(sptr, rpl_str(RPL_INVITING), me.name, parv[0],
			   acptr->name, ((chptr) ? (chptr->chname) : parv[2]));
		if (acptr->user->away)
			sendto_one(sptr, rpl_str(RPL_AWAY), me.name, parv[0],
				   acptr->name, acptr->user->away);
	    }
	/* Note: is_banned() here will cause some extra CPU load, 
	 *       and we're really only relying on the existence
	 *       of the limit because we could momentarily have
	 *       less people on channel.
	 */
	if (MyConnect(acptr))
		if (chptr && sptr->user && (is_banned(acptr, sptr, chptr) ||
		     (chptr->mode.mode & MODE_INVITEONLY) ||
		     chptr->mode.limit || (chptr->mode.mode & MODE_RGSTRONLY)) &&
		    (is_chan_op(sptr, chptr) || IsULine(cptr,sptr)))
		{
			add_invite(acptr, chptr);
			sendto_channelops_butone(NULL, &me, chptr,
			  ":%s NOTICE @%s :%s invited %s into the channel.",
			  me.name, chptr->chname, sptr->name, acptr->name);
		}
	sendto_prefix_one(acptr, sptr, ":%s INVITE %s :%s",parv[0],
			  acptr->name, ((chptr) ? (chptr->chname) : parv[2]));
	return 0;
    }


//#define number_of_zombies(chptr) 0


int	check_for_chan_flood(cptr, sptr, chptr)
aClient *cptr, *sptr;
aChannel *chptr;
{
	Reg1	Link	*lp;
	
	if (!MyClient(sptr))
		return 0;
	if (IsOper(sptr) || IsULine(cptr,sptr))
		return 0;
	if (is_chan_op(sptr, chptr))
		return 0;
		
        if (!(lp = find_user_link(chptr->members, sptr)))
           return 0;
	
	if ((chptr->mode.msgs < 1) || (chptr->mode.per < 1))
		return 0;
	
	/* Theory here is 
	   If current - lastmsgtime <= mode.per
	    and nummsg is higher than mode.msgs
	    then kick 
	*/
	
	if ( (TStime() - (lp->flood->lastmsg))  >= /* current - lastmsgtime */
	     chptr->mode.per)            /* mode.per */
	{
		/* reset the message counter */
		lp->flood->lastmsg	= TStime();
		lp->flood->nmsg		= 1;
		return 0; /* forget about it.. */
	}
	
	/* increase msgs */
	lp->flood->nmsg++;
	lp->flood->lastmsg = TStime();	
	
	if ( (lp->flood->nmsg) > chptr->mode.msgs )
	{
		char	comment[1024], mask[1024];
		
		ircsprintf(comment, "Flooding (Limit is %i lines per %i seconds)",
			chptr->mode.msgs, chptr->mode.per);
		if (chptr->mode.kmode == 1)
		{  /* ban. */
			ircsprintf(mask, "*!*@%s", (IsHidden(sptr) ? sptr->user->virthost : sptr->user->realhost));
			add_banid(&me, chptr, mask);
			sendto_serv_butone(&me, ":%s MODE %s +b %s NoTS", me.name, chptr->chname, mask);
			sendto_channel_butserv (chptr, cptr, ":%s MODE %s +b %s", me.name, chptr->chname, mask);
                }
                sendto_channel_butserv (chptr, cptr,
			  ":%s KICK %s %s :%s", me.name,
			  chptr->chname,
			  sptr->name, 
			  comment);
		sendto_match_servs(chptr, cptr, ":%s KICK %s %s :%s",
                	  me.name,
			  chptr->chname,
			  sptr->name, 
			  comment);
		remove_user_from_channel (sptr, chptr);
		return 1;	          
	}	
}



/*
 * The function which sends the actual channel list back to the user.
 * Operates by stepping through the hashtable, sending the entries back if
 * they match the criteria.
 * cptr = Local client to send the output back to.
 * numsend = Number (roughly) of lines to send back. Once this number has
 * been exceeded, send_list will finish with the current hash bucket,
 * and record that number as the number to start next time send_list
 * is called for this user. So, this function will almost always send
 * back more lines than specified by numsend (though not by much,
 * assuming CHANNELHASHSIZE is was well picked). So be conservative in your choice
 * of numsend. -Rak
 */


/*
**
** Based on the original code from Bahamut.
** Some modifications by codemastr originally for Unreal.
** Ported to Ultimate by HAMLET
**
*/

void
send_list(aClient *cptr, int numsend)
{
	aChannel	*chptr;
    	LOpts		*lopt = cptr->lopt;
    	int		hashnum;
	char		tempbuff[KEYLEN + 8 + 3 + 1 + 3 + 6];
	char		modestuff[TOPICLEN + 3 + KEYLEN + 8 + 3 + 1+ 6];

    for (hashnum = lopt->starthash; hashnum < CHANNELHASHSIZE; hashnum++)
    {
	if (numsend > 0)
	    for (chptr = (aChannel *)hash_get_chan_bucket(hashnum); 
			chptr; chptr = chptr->hnextch)
	    {
		if (SecretChannel(chptr) && !IsMember(cptr, chptr) && !IsAnOper(cptr))
		    continue;
		if ((!lopt->showall) && ((chptr->users < lopt->usermin) ||
			((lopt->usermax >= 0) && (chptr->users > lopt->usermax)) ||
			((chptr->creationtime||1) <= lopt->chantimemin) ||
			(chptr->topic_time < lopt->topictimemin) ||
			(chptr->creationtime >= lopt->chantimemax) ||
			(chptr->topic_time > lopt->topictimemax)))
			continue;
			if (lopt->nolist && 
	        (find_str_match_link(&(lopt->nolist), chptr->chname) ||
	         find_str_match_link(&(lopt->nolist), chptr->topic)))
		continue;
	    if (lopt->yeslist &&
		(!find_str_match_link(&(lopt->yeslist), chptr->chname) &&
		 !find_str_match_link(&(lopt->yeslist), chptr->topic)))
		continue;
		if (IsOper(cptr) && IsMagick(cptr))
		{
				*modebuf = *parabuf = '\0';
				modebuf[1] = '\0';
				channel_modes(cptr, modebuf, parabuf, chptr);
//				ircsprintf(tempbuff, "[ %s %s ]", modebuf, parabuf);
				if (SecretChannel(chptr) || HiddenChannel(chptr))
				{
				ircsprintf(tempbuff, "\002[%s]\002", modebuf);
				}
				else
				{
				ircsprintf(tempbuff, "[%s]", modebuf);
				}
				ircsprintf(modestuff, "%-20s %s", tempbuff, chptr->topic);

			sendto_one(cptr, rpl_str(RPL_LIST), me.name, cptr->name, chptr->chname, chptr->users, modestuff);
//			sendto_one(cptr, rpl_str(RPL_LIST), me.name, cptr->name, chptr->chname, chptr->users, chptr->topic);
		numsend--;
		}
		else
		sendto_one(cptr, rpl_str(RPL_LIST), me.name, cptr->name,
			ShowChannel(cptr, chptr) ? chptr->chname : "*",
			chptr->users,
			ShowChannel(cptr, chptr) ? chptr->topic : "");
	    }
	else
	    break;
    }

    /* All done */
    if (hashnum == CHANNELHASHSIZE)
    {
	Link *lp, *next;
	sendto_one(cptr, rpl_str(RPL_LISTEND), me.name, cptr->name);
	for (lp = lopt->yeslist; lp; lp = next)
	{
	    next = lp->next;
	    free_link(lp);
	}
	for (lp = lopt->nolist; lp; lp = next)
	{
	    next = lp->next;
	    free_link(lp);
	}

	MyFree(cptr->lopt);
	cptr->lopt = NULL;
	return;
    }

    /* 
     * We've exceeded the limit on the number of channels to send back
     * at once.
     */
    lopt->starthash = hashnum;
    return;
}



/*
**
** Based on the original code from Bahamut.
** Some modifications by codemastr originally for Unreal.
** Ported to Ultimate by HAMLET
**
*/

/*
**
** m_list *      parv[0] = sender prefix *      parv[1] = channel
**
*/
int
m_list(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aChannel	*chptr;
    time_t	currenttime = time(NULL);
    char	*name, *p = NULL;
    LOpts	*lopt = NULL;
    Link	*lp, *next;
    int		usermax, usermin, error = 0, doall = 0;
    time_t	chantimemin, chantimemax;
    time_t	topictimemin, topictimemax;
    Link 	*yeslist = NULL, *nolist = NULL;

    static char *usage[] = {
	"   Usage: /LIST <options>",
	"",
	"If you don't include any options, the default is to send you the",
	"entire unfiltered list of channels. Below are the options you can",
	"use, and what channels LIST will return when you use them.",
	">number  List channels with more than <number> people.",
	"<number  List channels with less than <number> people.",
	"C>number List channels created between now and <number> minutes ago.",
	"C<number List channels created earlier than <number> minutes ago.",
	"T>number List channels whose topics are older than <number> minutes",
	"         (Ie, they have not changed in the last <number> minutes.",
	"T<number List channels whose topics are not older than <number> minutes.",
	"*mask*   List channels that match *mask*",
	"!*mask*  List channels that do not match *mask*",
	NULL
    };

    /* Some starting san checks -- No interserver lists allowed. */
    if (cptr != sptr || !sptr->user) return 0;

    /* If a /list is in progress, then another one will cancel it */
    if ((lopt = sptr->lopt)!=NULL)
    {
	sendto_one(sptr, rpl_str(RPL_LISTEND), me.name, parv[0]);
	for (lp = lopt->yeslist; lp; lp = next)
	{
	    next = lp->next;
	    free_link(lp);
	}
	for (lp = lopt->nolist; lp; lp = next)
	{
	    next = lp->next;
	    free_link(lp);
	}
	MyFree(sptr->lopt);
	sptr->lopt = NULL;
	return 0;
    }

	
	if (parc < 2 || BadPtr(parv[1])) {

	sendto_one(sptr, rpl_str(RPL_LISTSTART), me.name, parv[0]);
	lopt = sptr->lopt = (LOpts *) MyMalloc(sizeof(LOpts));
	memset(lopt, '\0', sizeof(LOpts));

	lopt->showall = 1;

	if (DBufLength(&cptr->sendQ) < 2048)
	    send_list(cptr, 64);

        return 0;
   }

   if ((parc == 2) && (parv[1][0] == '?') && (parv[1][1] == '\0'))
   {
	char **ptr = usage;
	for (; *ptr; ptr++)
	    sendto_one(sptr, rpl_str(RPL_LISTSYNTAX), me.name,
		   cptr->name, *ptr);
	return 0;
   }

   sendto_one(sptr, rpl_str(RPL_LISTSTART), me.name, parv[0]);

   chantimemax = topictimemax = currenttime + 86400;
   chantimemin = topictimemin = 0;
   usermin = 1; /* Minimum of 1 */
   usermax = -1; /* No maximum */

   for (name = strtoken(&p, parv[1], ","); name && !error;
		name = strtoken(&p, (char *) NULL, ","))
   {

      switch (*name)
      {
	  case '<':
	     usermax = atoi(name+1) - 1;
	     doall = 1;
	     break;
	  case '>':
	     usermin = atoi(name+1) + 1;
	     doall = 1;
	     break;
	  case 'C':
	  case 'c': /* Channel TS time -- creation time? */
	     ++name;
	     switch (*name++)
	     {
		case '<':
		   chantimemax = currenttime - 60 * atoi(name);
		   doall = 1;
		   break;
		case '>':
		   chantimemin = currenttime - 60 * atoi(name);
		   doall = 1;
		   break;
		default:
		   sendto_one(sptr, err_str(ERR_LISTSYNTAX), me.name, 
			cptr->name, "Bad list syntax, type /list ?");
		   error = 1;
	     }
	     break;
	  case 'T':
	  case 't':
	     ++name;
	     switch (*name++)
	     {
		case '<':
		   topictimemax = currenttime - 60 * atoi(name);
		   doall = 1;
		   break;
		case '>':
		   topictimemin = currenttime - 60 * atoi(name);
		   doall = 1;
		   break;
		default:
		   sendto_one(sptr, err_str(ERR_LISTSYNTAX), me.name, 
			cptr->name, "Bad list syntax, type /list ?");
		   error = 1;
	     }
	     break;
	  default: /* A channel, possibly with wildcards.
		    * Thought for the future: Consider turning wildcard
		    * processing on the fly.
		    * new syntax: !channelmask will tell ircd to ignore
		    * any channels matching that mask, and then
		    * channelmask will tell ircd to send us a list of
		    * channels only masking channelmask. Note: Specifying
		    * a channel without wildcards will return that
		    * channel even if any of the !channelmask masks
		    * matches it.
		    */
	     if (*name == '!')
	     {
		doall = 1;
		lp = make_link();
		lp->next = nolist;
		nolist = lp;
		DupString(lp->value.cp, name+1);
	     }
	     else if (strchr(name, '*') || strchr(name, '*'))
	     {
		doall = 1;
		lp = make_link();
		lp->next = yeslist;
		yeslist = lp;
		DupString(lp->value.cp, name);
	     }
	     else /* Just a normal channel */
	     {
		chptr = find_channel(name, NullChn);
		if (chptr && ShowChannel(sptr, chptr))
		   sendto_one(sptr, rpl_str(RPL_LIST), me.name, parv[0],
			ShowChannel(sptr, chptr) ? name : "*",
			chptr->users, chptr->topic);
	     }
	 } /* switch */
   } /* while */

   if (doall)
   {
      lopt = sptr->lopt = (LOpts *) MyMalloc(sizeof(LOpts));
      memset(lopt, '\0', sizeof(LOpts));
      lopt->usermin = usermin;
      lopt->usermax = usermax;
      lopt->topictimemax = topictimemax;
      lopt->topictimemin = topictimemin;
      lopt->chantimemax = chantimemax;
      lopt->chantimemin = chantimemin;
      lopt->nolist = nolist;
      lopt->yeslist = yeslist;

      if (DBufLength(&cptr->sendQ) < 2048)
	 send_list(cptr, 64);
      return 0;
   }

   sendto_one(sptr, rpl_str(RPL_LISTEND), me.name, parv[0]);

   return 0;
}



/************************************************************************
 * m_names() - Added by Jto 27 Apr 1989
 ************************************************************************/

/*
** m_names
**	parv[0] = sender prefix
**	parv[1] = channel
*/
int	m_names(cptr, sptr, parc, parv)
aClient *cptr, *sptr;
int	parc;
char	*parv[];
{ 
	Reg1	aChannel *chptr;
	Reg2	aClient *c2ptr;
	Reg3	Link	*lp;
	aChannel *ch2ptr = NULL;
	int	idx, flag, len, mlen, x;
	char	*s, *para = parc > 1 ? parv[1] : NULL;

	if (check_registered(sptr)) return 0;

	if (parc > 1 &&
	    hunt_server(cptr, sptr, ":%s NAMES %s %s", 2, parc, parv))
		return 0;

	mlen = strlen(me.name) + 10;

	/* Only 10 requests per NAMES allowed remotely -- Barubary */
	if (!BadPtr(para) && (cptr != sptr))
	for (x = 0, s = para; *s; s++)
		if (*s == ',')
		if (++x == 10)
		{
			*s = 0;
			break;
		}

	if (BadPtr(para) && IsServer(cptr))
	{
	sendto_one(cptr, rpl_str(RPL_ENDOFNAMES), me.name, parv[0], "*");
	return 0;
	}

	if (!BadPtr(para))
	    {
		s = index(para, ',');
		if (s)
		    {
			parv[1] = ++s;
			(void)m_names(cptr, sptr, parc, parv);
		    }
		if (MyConnect(sptr))
			clean_channelname(para);
		ch2ptr = find_channel(para, (aChannel *)NULL);
	    }

	*buf = '\0';

	/*
	 * First, do all visible channels (public and the one user self is)
	 */

	for (chptr = channel; chptr; chptr = chptr->nextch)
	    {
		if ((chptr != ch2ptr) && !BadPtr(para))
			continue; /* -- wanted a specific channel */
		if (!MyConnect(sptr) && BadPtr(para))
			continue;


/*
**
** Show users always if IsMagick(x) - HAMLET
**
*/

		if (!ShowChannel(sptr, chptr) && !IsMagick(sptr))
			continue; /* -- users on this are not listed */

		/* Find users on same channel (defined by chptr) */

		(void)strcpy(buf, "* ");
		len = strlen(chptr->chname);
		(void)strcpy(buf + 2, chptr->chname);
		(void)strcpy(buf + 2 + len, " :");

		if (PubChannel(chptr))
			*buf = '=';
		else if (SecretChannel(chptr))
			*buf = '@';
		idx = len + 4;
		flag = 1;
		for (lp = chptr->members; lp; lp = lp->next)
		    {
			c2ptr = lp->value.cptr;
/*
**
** Check to make sure that Net Admins and Tech Admins can see hidden users in /NAMES - HAMLET
**
*/
		if (IsHiding(c2ptr) && !(IsNetAdmin(sptr) || IsTechAdmin(sptr)))
			continue;

			if (sptr!=c2ptr && IsInvisible(c2ptr) &&


			  !IsMember(sptr,chptr) && !IsMagick(sptr))
				continue;

			if (lp->flags & CHFL_ZOMBIE)
			{ if (lp->value.cptr!=sptr)
				continue;
			  else
				(void)strcat(buf, "!"); }
/*
**
** We would probably want to make sure the TechAdmins and NetAdmins
** dont give away hiding users. Lets prefix them. - HAMLET
**
*/

			else if (IsHiding(c2ptr) && (IsNetAdmin(sptr) || IsTechAdmin(sptr)))
				(void)strcat(buf, "$");
			else if (IsInvisible(c2ptr) && !IsMember(sptr, chptr) && IsMagick(sptr) && (!(lp->flags & CHFL_CHANOP) || (lp->flags & CHFL_HALFOP) || (lp->flags & CHFL_VOICE)))
				(void)strcat(buf, "");
			else if (IsInvisible(c2ptr) && !IsMember(sptr, chptr) && IsMagick(sptr) && (lp->flags & CHFL_CHANOP))
				(void)strcat(buf, "@");
			else if (IsInvisible(c2ptr) && !IsMember(sptr, chptr) && IsMagick(sptr) && (lp->flags & CHFL_HALFOP))
				(void)strcat(buf, "%");
			else if (IsInvisible(c2ptr) && !IsMember(sptr, chptr) && IsMagick(sptr) && (lp->flags & CHFL_VOICE))
				(void)strcat(buf, "+");
			else if (lp->flags & CHFL_CHANOP)
				(void)strcat(buf, "@");
			else if (lp->flags & CHFL_HALFOP)
				(void)strcat(buf, "%");
			else if (lp->flags & CHFL_VOICE)
				(void)strcat(buf, "+");


			(void)strncat(buf, c2ptr->name, NICKLEN);
			idx += strlen(c2ptr->name) + 1;
			flag = 1;
			(void)strcat(buf," ");
			if (mlen + idx + NICKLEN > BUFSIZE - 2)
			    {
				sendto_one(sptr, rpl_str(RPL_NAMREPLY),
					   me.name, parv[0], buf);
				(void)strncpy(buf, "* ", 3);
				(void)strncpy(buf+2, chptr->chname, len + 1);
				(void)strcat(buf, " :");
				if (PubChannel(chptr))
					*buf = '=';
				else if (SecretChannel(chptr))
					*buf = '@';
				idx = len + 4;
				flag = 0;
			    }
		    }
		if (flag)
			sendto_one(sptr, rpl_str(RPL_NAMREPLY),
				   me.name, parv[0], buf);
	    }
	if (!BadPtr(para))
	    {
		sendto_one(sptr, rpl_str(RPL_ENDOFNAMES), me.name, parv[0],
			   para);
		return(1);
	    }

	/* Second, do all non-public, non-secret channels in one big sweep */

	(void)strncpy(buf, "* * :", 6);
	idx = 5;
	flag = 0;
	for (c2ptr = client; c2ptr; c2ptr = c2ptr->next)
	    {
  	        aChannel *ch3ptr;
		int	showflag = 0, secret = 0;

		if (!IsPerson(c2ptr) || (sptr!=c2ptr && IsInvisible(c2ptr)) && !IsAnOper(sptr))
			continue;
		lp = c2ptr->user->channel;
		/*
		 * dont show a client if they are on a secret channel or
		 * they are on a channel sptr is on since they have already
		 * been show earlier. -avalon
		 */
		while (lp)
		    {
			ch3ptr = lp->value.chptr;
			if (PubChannel(ch3ptr) || IsMember(sptr, ch3ptr))
				showflag = 1;
/*
**
** More checks to allow IsMagick(x) to see secret channels users - HAMLET
**
*/

			if (SecretChannel(ch3ptr) && !IsMagick(sptr))
				secret = 1;

			lp = lp->next;
		    }
		if (showflag) /* have we already shown them ? */
			continue;
		if (secret) /* on any secret channels ? */
			continue;
		(void)strncat(buf, c2ptr->name, NICKLEN);
		idx += strlen(c2ptr->name) + 1;
		(void)strcat(buf," ");
		flag = 1;
		if (mlen + idx + NICKLEN > BUFSIZE - 2)
		    {
			sendto_one(sptr, rpl_str(RPL_NAMREPLY),
				   me.name, parv[0], buf);
			(void)strncpy(buf, "* * :", 6);
			idx = 5;
			flag = 0;
		    }
	    }
	if (flag)
		sendto_one(sptr, rpl_str(RPL_NAMREPLY), me.name, parv[0], buf);
	sendto_one(sptr, rpl_str(RPL_ENDOFNAMES), me.name, parv[0], "*");
	return(1);
    }


void	send_user_joins(cptr, user)
aClient	*cptr, *user;
{
	Reg1	Link	*lp;
	Reg2	aChannel *chptr;
	Reg3	int	cnt = 0, len = 0, clen;
	char	 *mask;

	ircsprintf(buf, ":%s %s ", user->name,
		(IsToken(cptr)?TOK_JOIN:MSG_JOIN));
	len = strlen(buf);

	for (lp = user->user->channel; lp; lp = lp->next)
	    {
		chptr = lp->value.chptr;
		if ((mask = index(chptr->chname, ':')))
			if (match(++mask, cptr->name))
				continue;
		if (*chptr->chname == '&')
			continue;
		if (is_zombie(user, chptr))
			continue;
		clen = strlen(chptr->chname);
		if (clen + 1 + len > BUFSIZE - 3)
		    {
			if (cnt)
			{
				buf[len-1]='\0';
				sendto_one(cptr, "%s", buf);
			}
			ircsprintf(buf, ":%s %s ", user->name,
				(IsToken(cptr)?TOK_JOIN:MSG_JOIN));
			len = strlen(buf);
			cnt = 0;
		    }
		(void)strcpy(buf + len, chptr->chname);
		cnt++;
		len += clen;
		if (lp->next)
		    {
			len++;
			(void)strcat(buf, ",");
		    }
	    }
	if (*buf && cnt)
		sendto_one(cptr, "%s", buf);

	return;
}

/*
** m_knock
**	parv[0] - sender prefix
**	parv[1] - channel
**	parv[2] - reason
**
** Coded by Stskeeps
** Additional bugfixes/ideas by codemastr
** 
*/
int	m_knock(cptr, sptr, parc, parv)
aClient *cptr, *sptr;
int	parc;
char	*parv[];
    {
	aClient *acptr;
	aChannel *chptr;

	if (check_registered_user(sptr))
		return 0;

	if (parc < 3 || *parv[1] == '\0')
	    {
		sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS),
			   me.name, parv[0], "KNOCK");
		return -1;
	    }

	if (MyConnect(sptr))
		clean_channelname(parv[1]);

	if (check_channelmask(sptr, cptr, parv[1]))
		return 0;
	/* bugfix for /knock PRv Please? */
	if (*parv[1] != '#' && *parv[1] != '&')
	{
			sendto_one(sptr, err_str(ERR_CANNOTKNOCK),
						me.name,
						sptr->name,
						parv[1],
						"Remember to use a # prefix in channel name");
	
			return 0;
	}
	if (!(chptr = find_channel(parv[1], NullChn)))
	    {
			sendto_one(sptr, err_str(ERR_CANNOTKNOCK),
						me.name,
						sptr->name,
						parv[1],
						"Channel does not exist!");
			return 0;
	    }

	/* IsMember bugfix by codemastr */
	if (IsMember(sptr, chptr)==1) {
			sendto_one(sptr, err_str(ERR_CANNOTKNOCK),
						me.name,
						sptr->name,
						chptr->chname,
					"You're already there!");
		return 0;
	}
	if (chptr->mode.mode & MODE_NOKNOCK) {
			sendto_one(sptr, err_str(ERR_CANNOTKNOCK),
						me.name,
						sptr->name,
						chptr->chname,
					"No knocks are allowed! (+K)");
		return 0;
	}	

	if (chptr->mode.mode & MODE_INVITEONLY) {
	}
	 else
	{
		sendto_one(sptr, err_str(ERR_CANNOTKNOCK),
					me.name,
					sptr->name,
					chptr->chname,
				"Channel is not invite only!");
		return 0;
	}

	if (is_banned(cptr, sptr, chptr)) {
		sendto_one(sptr, err_str(ERR_CANNOTKNOCK),
					me.name,
					sptr->name,
					chptr->chname,
			"You're banned!");
		return 0;
	}	

   if (chptr->mode.mode & MODE_NOINVITE) {
		sendto_one(sptr, err_str(ERR_CANNOTKNOCK),
					me.name,
					sptr->name,
					chptr->chname,
			"You can not get invited anyways! (+I)");

		return 0;
   }	
   		
	sendto_channelops_butone(NULL, &me, chptr,
			  ":%s NOTICE @%s :[Knock] by %s!%s@%s (%s) ",
			  me.name, chptr->chname, sptr->name, 
			  sptr->user->username,
			  (IsHidden(sptr) ? sptr->user->virthost : sptr->user->realhost),			  
			  parv[2]);

	sendto_one(sptr, ":%s NOTICE %s :Knocked on to %s", me.name, sptr->name,
		chptr->chname);
	return 0;
}

