/*   UltimateIRCd - Ultimate Internet Relay Chat Daemon, src/s_addline.c
**
**   Written for UltimateIRCd by ShadowRealm Creations.
**   Copyright (C) 1997-2000 Infomedia Inc.
**
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 1, or (at your option)
**   any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/un.h>
#include "true.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "userload.h"
#include <sys/stat.h>
#include <fcntl.h>
#include <utmp.h>
#include "h.h"


extern char *configfile;
extern char *elinefile;
extern char *klinefile;

extern u_long inet_addr(const char *cp);



/*
 * m_addcnline
 * m_addcnline - This function adds c/n lines to the ircd.conf
 * checks that the user have the e flag in the O line then checks
 * the first parameter and if it's valid them passes the parametres to the
 * correct function.
 *  parv[0] = real hostname
 *  parv[1] = password
 *  parv[2] = network name
 */
int m_addcnline(aClient *cptr, aClient *sptr, int parc, char **parv)
{
  FILE *fp;
  char usrstr[300];
  time_t cur = time (NULL);

  if (check_registered(sptr))
    return -1;

  if (!MyClient(sptr) || !OpCanEDConf(sptr))
  {
    sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
      return -1;
  }

  if (parc < 6)
  {
      sendto_one (sptr, ":%s NOTICE %s :Help on ADDCNLINE", me.name, parv[0]);
    sendto_one(sptr,
               ":%s NOTICE %s :Syntax: /ADDCNLINE <ip/hostname> <password>"
               " <servername> <port|*> <class|30>", me.name, parv[0]);
    return -1;
  }

  /*
  if (strchr (parv[1], '*') || strchr (parv[2], '*') || strchr (parv[3], '*') || strchr (parv[5], '*'))
  {
    sendto_one(sptr, ":%s NOTICE %s :You can't use WildCards!",
               me.name, parv[0]);
    return -1;
  }
  */

  sendto_one(sptr, "NOTICE %s :Adding C/N lines", parv[0]);
  fp = fopen(configfile, "a");
  fprintf(fp, "\n");
  fprintf(fp, "#CNLines# Added by %s (%s@%s) on %s",
           parv[0], sptr->user->username, sptr->sockhost, ctime (&cur));
  ircsprintf(usrstr, "C:%s:%s:%s:%s:%s\n", parv[1], parv[2], parv[3],
		parv[4], parv[5]);
  fprintf(fp, "%s", usrstr);
  fprintf(fp, "N:%s:%s:%s:*:%s\n", parv[1], parv[2], parv[3], parv[5]);
  fclose(fp);


  if (!rehash (cptr, sptr, 0))
    {
      sendto_one (sptr, ":%s NOTICE %s :C/N now active for %s (%s) Class %s",
                  me.name, parv[0], parv[1], parv[3], parv[5]);

sendto_serv_butone(&me, ":%s NETG :%s has added an active C/N line set for %s (%s) Class %s", me.name, parv[0], parv[1], parv[3], parv[5]);

sendto_netglobal("from %s : %s has added an active C/N line set for %s (%s) Class %s", me.name, parv[0], parv[1], parv[3], parv[5]);

  }
  else
    {
      sendto_one (sptr, ":%s NOTICE %s :ERROR: rehash() failed!",
                  me.name, parv[0]);
      return -1;
    }
  return 0;
}




/*
**
** m_delcnline
**
*/
int m_delcnline(aClient *cptr, aClient *sptr, int parc, char **parv)
{
  aConfItem *aconf;
  FILE *in, *out;
  char buf[256], temppath[256];
  ircsprintf(temppath, "%s.tmp", configfile);

  if (check_registered (sptr))
    return -1;

  if (!OpCanEDConf(sptr) || !MyClient (sptr))
    {
      sendto_one (sptr, err_str (ERR_NOPRIVILEGES), me.name, parv[0]);
      return -1;
    }

  if (parc < 2)
    {
      sendto_one (sptr, ":%s NOTICE %s :Help on DELCNLINE", me.name, parv[0]);
      sendto_one (sptr, ":%s NOTICE %s :Syntax /DELCNLINE <servername>", me.name,
                  parv[0]);
      return -1;
    }

  for (aconf = conf; aconf; aconf = aconf->next)
    if (aconf->status == CONF_CONNECT_SERVER && !match(parv[1], aconf->name))
	break;
  if (!aconf)
    {
      sendto_one(sptr, ":%s NOTICE %s :DELCNLINE: Host %s not listed in %s",
                  me.name, parv[0], parv[1], configfile);
      return -1;
    }

  in = fopen (configfile, "r");
  out = fopen (temppath, "w");
  while (fgets (buf, sizeof (buf), in))
    {
      if (!strstr (buf, parv[1]))
        fputs (buf, out);
    }
  fclose (in);
  fclose (out);
  if (rename (temppath, configfile) == -1)
    {
sendto_one (sptr, ":%s NOTICE %s :ERROR: rename() failed!", me.name, parv[0]);
      return -1;
    }

  if (!rehash (cptr, sptr, 0))
    {
      sendto_one (sptr, "NOTICE %s :%s has been removed",
                  parv[0], parv[1]);

sendto_serv_butone(&me, ":%s NETG :%s removed C/N line set for %s", me.name, parv[0], parv[1]);

sendto_netglobal("from %s : %s removed C/N line set for %s", me.name, parv[0], parv[1]);

    }
  else
    {
      sendto_one (sptr, ":%s NOTICE %s :ERROR: rehash() failed!",
                  me.name, parv[0]);
      return -1;
    }
  return 0;
}


/*
**
** m_addoper
**
*/

int m_addoper(aClient *cptr, aClient *sptr, int parc, char **parv)
{
  char usrstr[300], *lg = '\0';
  time_t cur = time (NULL);
  FILE *fp;
#ifdef CRYPT_OPER_PASSWORD
  static char saltChars[] = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRS"
                            "TUVWXYZ0123456789./";
  char salt[3];
#endif

  if (check_registered (sptr))
    return -1;

  if (!OpCanEDConf(sptr) || !MyClient(sptr))
    {
      sendto_one (sptr, err_str (ERR_NOPRIVILEGES), me.name, parv[0]);
      return -1;
    }

  if (parc < 6)
    {
      sendto_one (sptr, ":%s NOTICE %s :Help on ADDOPER", me.name,
                  parv[0]);
      sendto_one (sptr, ":%s NOTICE %s :Syntax: /ADDOPER <operhost>"
                        " <password> <opernick> <operflags> <class|10>",
                  me.name, parv[0]);
      return -1;
    }


  if ((strlen (parv[2]) < 5))
    {
      sendto_one (sptr, ":%s NOTICE %s :The password %s is an insecure "
                        "password, Please try again with a password with"
                        " at least 5 characters, including alpha-numeric, "
                        "upper and lower case characters",
                        me.name, parv[0], parv[3]);
      return -1;
    }

#ifdef CRYPT_OPER_PASSWORD
  srandom (time (0));           /* may not be the BEST salt, but its close */
  salt[0] = saltChars[random () % 64];
  salt[1] = saltChars[random () % 64];
  salt[2] = 0;
  if ((!strchr (saltChars, salt[0])) || (!strchr (saltChars, salt[1])))
    sendto_one (sptr, ":%s NOTICE %s: illegal salt %s",
                me.name, parv[0], salt);
#endif

strlower(parv[3]);

  sendto_one (sptr, ":%s NOTICE %s :Adding O-line for %s (%s)",
              me.name, parv[0], parv[3], parv[1]);
  fp = fopen (configfile, "a");
  fprintf(fp, "\n");
  fprintf (fp, "#O-Line# Added by %s (%s@%s) on %s",
           parv[0], sptr->user->username, sptr->sockhost, ctime (&cur));
#ifdef CRYPT_OPER_PASSWORD
  ircsprintf(usrstr, "O:%s:%s:%s:%s:%s\n", parv[1], crypt(parv[2], salt),
              parv[3], parv[4], parv[5]);
#else
  ircsprintf(usrstr, "O:%s:%s:%s:%s:%s\n", parv[1], parv[2],
			parv[3], parv[4], parv[5]);
#endif
  fprintf (fp, "%s", usrstr);
  fclose (fp);
  if (!rehash (cptr, sptr, 0))
    {
      sendto_one (sptr, ":%s NOTICE %s :O-line now active for %s (%s)",
                  me.name, parv[0], parv[3], parv[1]);

sendto_serv_butone(&me, ":%s NETG :%s has added an active O-line for %s (%s) with flags %s", me.name, parv[0], parv[3], parv[1], parv[4]);

sendto_netglobal("from %s : %s has added an active O-line for %s (%s) with flags %s", me.name, parv[0], parv[3], parv[1], parv[4]);

    }
  else
    {
sendto_one (sptr, ":%s NOTICE %s :ERROR: rehash() failed!", me.name, parv[0]);
      return -1;
    }
  return 0;
}

/*
**
** m_deloper
**
*/


int m_deloper(aClient *cptr, aClient *sptr, int parc, char **parv)
{
  aConfItem *aconf;
  FILE *in, *out;
  char buf[256], temppath[256];

  ircsprintf(temppath, "%s.tmp", configfile);

  if (check_registered (sptr))
    return -1;

  if (!OpCanEDConf(sptr) || !MyClient (sptr))
    {
      sendto_one (sptr, err_str (ERR_NOPRIVILEGES), me.name, parv[0]);
      return -1;
    }

  if (parc < 2)
    {
      sendto_one (sptr, ":%s NOTICE %s :Help on DELOPER", me.name, parv[0]);
      sendto_one (sptr, ":%s NOTICE %s :Syntax /DELOPER <opernick>", me.name,
                  parv[0]);
      return -1;
    }
strlower(parv[1]);

  if (!(aconf = find_conf_name (parv[1], CONF_OPS)))
    {
      sendto_one (sptr, ":%s NOTICE %s :DELOPER: Specified oper %s not listed"
                        " in %s", me.name, parv[0], parv[1], configfile);
      return -1;
    }

  in = fopen (configfile, "r");
  out = fopen (temppath, "w");
  while (fgets (buf, sizeof (buf), in))
    if (!strstr (buf, parv[1]))
      fputs (buf, out);

  fclose (in);
  fclose (out);
  if (rename (temppath, configfile) == -1)
    {
sendto_one (sptr, ":%s NOTICE %s :ERROR: rename() failed!", me.name, parv[0]);

      return -1;
    }
  if (!rehash (cptr, sptr, 0))
    {
      sendto_one (sptr, "NOTICE %s :%s's O-Line has been removed", parv[0],
                  parv[1]);

sendto_serv_butone(&me, ":%s NETG :%s is removing O-Line for %s", me.name, parv[0], parv[1]);

sendto_netglobal("from %s : %s is removing O-Line for %s", me.name, parv[0], parv[1]);

    }
  else
    {
sendto_one (sptr, ":%s NOTICE %s :ERROR: rehash() failed!", me.name, parv[0]);

      return -1;
    }
  return 0;
}


/*
**
** m_addqline
**
*/

int m_addqline(aClient *cptr, aClient *sptr, int parc, char **parv)
{
  FILE *fp;
  char usrstr[300];
  time_t cur = time (NULL);

  if (check_registered (sptr))
    return -1;

  if (!OpCanEDConf(sptr) || !MyClient(sptr))
    {
      sendto_one (sptr, err_str (ERR_NOPRIVILEGES), me.name, parv[0]);
      return -1;
    }

  if (parc < 3)
    {
      sendto_one (sptr, ":%s NOTICE %s :Help on ADDQLINE",
                  me.name, parv[0]);
      sendto_one (sptr, ":%s NOTICE %s :Syntax: /ADDQLINE <qnick> <reason>",
                  me.name, parv[0]);
      return -1;
    }

  if (mycmp(parv[0], parv[1]) == 0) {
	sendto_one(sptr, ":%s NOTICE %s :This is a wakeup call. Think yourself a little about.", me.name, parv[0]);

sendto_serv_butone(&me, ":%s NETG :%s Just tried to Q-Line himself.", me.name, parv[0]);

sendto_netglobal("from %s : %s Just Tried to Q-Line himself", me.name, parv[0], parv[1]);

	return -1;
}
  sendto_one (sptr, ":%s NOTICE %s :Adding Q-line for nick %s reason: %s",
              me.name, parv[0], parv[1], parv[2]);
  fp = fopen (configfile, "a");
  fprintf(fp, "\n");
  fprintf (fp, "#Q-Line# Added by %s (%s@%s) on %s",
           parv[0], sptr->user->username, sptr->sockhost, ctime (&cur));
  ircsprintf(usrstr, "Q::%s:%s\n", parv[2], parv[1]);
  fprintf (fp, "%s", usrstr);
  fclose (fp);
  if (!rehash (cptr, sptr, 0))
    {
      sendto_one (sptr, ":%s NOTICE %s :Q-LINE added for %s", me.name, parv[0],
                  parv[1]);

sendto_serv_butone(&me, ":%s NETG :%s is adding a Q-Line for nick %s reason: %s", me.name, parv[0], parv[1], parv[2]);

sendto_netglobal("from %s : %s is adding a Q-Line for nick %s reason: %s", me.name, parv[0], parv[1], parv[2]);

    }
  else
    {
sendto_one (sptr, ":%s NOTICE %s :ERROR: rehash() failed!", me.name, parv[0]);


      return -1;
    }
  return 0;
}


/*
**
** m_delqline
**
*/

int m_delqline(aClient *cptr, aClient *sptr, int parc, char **parv)
{
  aConfItem *aconf;
  FILE *in, *out;
  char buf[256], temppath[256];

  ircsprintf(temppath, "%s.tmp", configfile);

  if (check_registered (sptr))
    return -1;

  if (!OpCanEDConf(sptr) || !MyClient (sptr))
    {
      sendto_one (sptr, err_str (ERR_NOPRIVILEGES), me.name, parv[0]);
      return -1;
    }

  if (parc < 2)
    {
      sendto_one (sptr, ":%s NOTICE %s :Help on DELQLINE", me.name, parv[0]);
      sendto_one (sptr, ":%s NOTICE %s :Syntax /DELQLINE <qnick>", me.name,
                  parv[0]);
      return -1;
    }
  
  if (aconf->status == CONF_QUARANTINED_SERVER && !match(parv[1], aconf->name))
    {
      sendto_one (sptr, ":%s NOTICE %s :DELQLINE: Specified nick %s not listed"
                        " in %s",
                  me.name, parv[0], parv[1], configfile);
      return 0;
    }
  in = fopen (configfile, "r");
  out = fopen (temppath, "w");
  while (fgets (buf, sizeof (buf), in))
    if (!strstr (buf, parv[1]))
      fputs (buf, out);

  fclose (in);
  fclose (out);
  if (rename (temppath, configfile) == -1)
    {
sendto_one (sptr, ":%s NOTICE %s :ERROR: rename() failed!", me.name, parv[0]);

      return -1;
    }
  if (!rehash (cptr, sptr, 0))
    {
      sendto_one (sptr, "NOTICE %s :Q-Line for %s has been removed",
                  parv[0], parv[1]);

sendto_serv_butone(&me, ":%s NETG :%s have removed Q-line for %s", me.name, parv[0], parv[1]);

sendto_netglobal("from %s : %s have removed Q-line for %s", me.name, parv[0], parv[1]);

    }
  else
    {
sendto_one (sptr, ":%s NOTICE %s :ERROR: rehash() failed!", me.name, parv[0]);

      return -1;
    }
  return 0;
}

/*
**
** m_addhub
**
*/

int m_addhub(aClient *cptr, aClient *sptr, int parc, char **parv)
{
  FILE *fp;
  char usrstr[300];
  time_t cur = time (NULL);

  if (check_registered (sptr))
    return -1;

  if (!OpCanEDConf(sptr) || !MyClient (sptr))
    {
      sendto_one (sptr, err_str (ERR_NOPRIVILEGES), me.name, parv[0]);
      return -1;
    }

  if (parc < 2)
    {
      sendto_one (sptr, ":%s NOTICE %s :Help on ADDHUB", me.name, parv[0]);
      sendto_one (sptr, ":%s NOTICE %s :Syntax: /ADDHUB <server>", me.name,
                  parv[0]);
      return -1;
    }

  sendto_one (sptr, "NOTICE %s :Adding H-line for %s", parv[0], parv[1]);
  fp = fopen (configfile, "a");
  fprintf(fp, "\n");
  fprintf (fp, "#H-Line# Added by %s (%s@%s) on %s",
           parv[0], sptr->user->username, sptr->sockhost, ctime (&cur));
  ircsprintf(usrstr, "H:*::%s\n", parv[1]);
  fprintf (fp, "%s", usrstr);
  fclose (fp);
  if (!rehash (cptr, sptr, 0))
    {
      sendto_one (sptr, ":%s NOTICE %s :H-line added for %s",
                  me.name, parv[0], parv[1]);

sendto_serv_butone(&me, ":%s NETG :%s has added an active H-line for %s", me.name, parv[0], parv[1]);

sendto_netglobal("from %s : %s has added an active H-line for %s", me.name, parv[0], parv[1]);
    }
  else
    {
sendto_one (sptr, ":%s NOTICE %s :ERROR: rehash() failed!", me.name, parv[0]);

      return -1;
    }
  return 0;
}


/*
**
** m_delhub
**
*/
int m_delhub(aClient *cptr, aClient *sptr, int parc, char **parv)
{
  aConfItem *aconf;
  FILE *in, *out;
  char buf[256], temppath[256];

  ircsprintf(temppath, "%s.tmp", configfile);

  if (check_registered (sptr))
    return -1;

  if (!OpCanEDConf (sptr) || !MyClient (sptr))
    {
      sendto_one (sptr, err_str (ERR_NOPRIVILEGES), me.name, parv[0]);
      return -1;
    }

  if (parc < 2)
    {
      sendto_one (sptr, ":%s NOTICE %s :Help on DELHUB", me.name, parv[0]);
      sendto_one (sptr, ":%s NOTICE %s :Syntax /DELHUB <server>", me.name,
                  parv[0]);
      return -1;
    }
  
  if (aconf->status == CONF_HUB && !match(parv[1], aconf->name))
    {
      sendto_one (sptr, ":%s NOTICE %s :DELHLINE: Specified server %s not listed"
                        " in %s",
                  me.name, parv[0], parv[1], configfile);
      return 0;
    }
  in = fopen (configfile, "r");
  out = fopen (temppath, "w");
  while (fgets (buf, sizeof (buf), in))
    if (!strstr (buf, parv[1]))
      fputs (buf, out);

  fclose (in);
  fclose (out);
  if (rename (temppath, configfile) == -1)
    {
sendto_one (sptr, ":%s NOTICE %s :ERROR: rehash() failed!", me.name, parv[0]);
      return -1;
    }
  if (!rehash (cptr, sptr, 0))
    {
      sendto_one (sptr, "NOTICE %s :H-Line for %s has been removed",
                  parv[0], parv[1]);

sendto_serv_butone(&me, ":%s NETG :%s have removed H-line for %s", me.name, parv[0], parv[1]);

sendto_netglobal("from %s : %s have removed H-line for %s", me.name, parv[0], parv[1]);

    }
  else
    {
sendto_one (sptr, ":%s NOTICE %s :ERROR: rehash() failed!", me.name, parv[0]);

      return -1;
    }
  return 0;
}



/*
**
** m_adduline
** Added on November 11th by HAMLET
**
*/

int m_adduline(aClient *cptr, aClient *sptr, int parc, char **parv)
{
  FILE *fp;
  char usrstr[300];
  time_t cur = time (NULL);

  if (check_registered (sptr))
    return -1;

  if (!OpCanEDConf(sptr) || !MyClient (sptr))
    {
      sendto_one (sptr, err_str (ERR_NOPRIVILEGES), me.name, parv[0]);
      return -1;
    }

  if (parc < 2)
    {
      sendto_one (sptr, ":%s NOTICE %s :Help on ADDULINE", me.name, parv[0]);
      sendto_one (sptr, ":%s NOTICE %s :Syntax: /ADDULINE <server>", me.name,
                  parv[0]);
      return -1;
    }

  sendto_one (sptr, "NOTICE %s :Adding U-line for %s", parv[0], parv[1]);
  fp = fopen (configfile, "a");
  fprintf(fp, "\n");
  fprintf (fp, "#U-Line# Added by %s (%s@%s) on %s",
           parv[0], sptr->user->username, sptr->sockhost, ctime (&cur));
  ircsprintf(usrstr, "U:%s:*:*\n", parv[1]);
  fprintf (fp, "%s", usrstr);
  fclose (fp);
  if (!rehash (cptr, sptr, 0))
    {
      sendto_one (sptr, ":%s NOTICE %s :U-line added for %s",
                  me.name, parv[0], parv[1]);

sendto_serv_butone(&me, ":%s NETG :%s has added an active U-line for %s", me.name, parv[0], parv[1]);

sendto_netglobal("from %s : %s has added an active U-line for %s", me.name, parv[0], parv[1]);
    }
  else
    {
sendto_one (sptr, ":%s NOTICE %s :ERROR: rehash() failed!", me.name, parv[0]);

      return -1;
    }
  return 0;
}


/*
**
** m_deluline
** Added on November 11th by HAMLET
**
*/
int m_deluline(aClient *cptr, aClient *sptr, int parc, char **parv)
{
  aConfItem *aconf;
  FILE *in, *out;
  char buf[256], temppath[256];

  ircsprintf(temppath, "%s.tmp", configfile);

  if (check_registered (sptr))
    return -1;

  if (!OpCanEDConf (sptr) || !MyClient (sptr))
    {
      sendto_one (sptr, err_str (ERR_NOPRIVILEGES), me.name, parv[0]);
      return -1;
    }

  if (parc < 2)
    {
      sendto_one (sptr, ":%s NOTICE %s :Help on DELULINE", me.name, parv[0]);
      sendto_one (sptr, ":%s NOTICE %s :Syntax /DELULINE <server>", me.name,
                  parv[0]);
      return -1;
    }
  
  if (aconf->status == CONF_UWORLD && !match(parv[1], aconf->name))
    {
      sendto_one (sptr, ":%s NOTICE %s :DELULINE: Specified server %s not listed"
                        " in %s",
                  me.name, parv[0], parv[1], configfile);
      return 0;
    }
  in = fopen (configfile, "r");
  out = fopen (temppath, "w");
  while (fgets (buf, sizeof (buf), in))
    if (!strstr (buf, parv[1]))
      fputs (buf, out);

  fclose (in);
  fclose (out);
  if (rename (temppath, configfile) == -1)
    {
sendto_one (sptr, ":%s NOTICE %s :ERROR: rehash() failed!", me.name, parv[0]);
      return -1;
    }
  if (!rehash (cptr, sptr, 0))
    {
      sendto_one (sptr, "NOTICE %s :U-Line for %s has been removed",
                  parv[0], parv[1]);

sendto_serv_butone(&me, ":%s NETG :%s have removed U-line for %s", me.name, parv[0], parv[1]);

sendto_netglobal("from %s : %s have removed U-line for %s", me.name, parv[0], parv[1]);

    }
  else
    {
sendto_one (sptr, ":%s NOTICE %s :ERROR: rehash() failed!", me.name, parv[0]);

      return -1;
    }
  return 0;
}

